# LSB Test Execution Framework
# Journal & report module
#
# Copyright (C) 2007 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF with some parts
# derived from lsb-autotest written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Journal;

use strict;

use Common;
use Packages;
use IO::File;

# Export symbols
require Exporter;
@Journal::ISA = qw(Exporter);
push @Journal::EXPORT, qw(
    &parse_tet_journal &parse_tcl_journal &parse_python_journal 
    &parse_simple_journal &parse_raptor_journal 
    &generate_report
    &pack_and_send_output
);


# HTML colors used in the report.
my $dark_gray = "rgb(64, 64, 64)";
my $gray = "rgb(200, 200, 200)";
my $neutral_gray = "rgb(230, 230, 230)";
my $bright_gray = "rgb(240, 240, 240)";
my $fluorescent_green = "rgb(64, 255, 64)";
my $fluorescent_yellow = "rgb(255, 255, 64)";
my $fluorescent_red = "rgb(255, 96, 96)";
my $pale_blue = "rgb(220, 220, 240)";
my $dark_blue = "rgb(0, 0, 128)";
my $dark_red = "rgb(128, 0, 0)";
my $red = "rgb(255, 0, 0)";
my $green = "rgb(0, 160, 0)";
my $blue = "rgb(0, 0, 255)";

# Statistics for each test
my %stats_failures_reported = ();
my %stats_warnings_reported = ();
my %stats_fips_reported = ();
my %stats_notices_reported = ();
my %stats_tests_total = ();
my %stats_tests_passed = ();
my %stats_tests_skipped = ();
my %stats_tests_quelled = ();
my %stats_tests_expected = ();
my %stats_tests_skips_unexpected = ();

# Summary statistics
my $tests_total = 0;
my $tests_passed = 0;
my $tests_quelled = 0;
my $tests_skipped = 0;

my $tests_cancelled = 0;
my $failures_reported = 0;
my $overall_verdict = "";

# The problem database, hashed by the "test_name test_point" strings.
my %problem_database;

# Object representing a test being processed.
my $test_object;

# The final exit code, which is 0 if all tests have passed, or 1 if some
# failures were detected.
my $exit_code = 0;

# The report file.
my $report;

# The single-line summary file.
my $summary_line;

# This variable is set to 1 if part of the problem table was already printed
# (at least the header), false if not (this implies that no problems were
# found so far).
my $problem_table_printed = 0;

# A monotonously increasing, unique identifier of a message box.
my $msgbox_id = 1;

# Copy of the input line that was processed last.
my $line = "";

#----------------------------------------------------------------------------

# The journal file being processed.
my $journal_file;
my $journal_name;

# The title of the journal being processed.
my $journal_title = "";
my @journal_titles = ();

my %journal_group = ();

# The HTML anchors used for marking this journal in the resulting report.
my %journal_anchors = ();

# Reports for each test suite
my %reports = ();  # $reports{$journal_title} = $html_report_text

sub report {
    $_ = shift;
    $reports{$journal_title} .= $_;
}

my %test_versions = (); # $test_versions{$journal_title} = <test_version>
#----------------------------------------------------------------------------

# Map of TET result codes to human-readable severity names.
my %result_code_map = (
    0, "Passed",
    1, "Failed",
    2, "Unresolved",
    3, "Unused",
    4, "Unsupported",
    5, "Untested",
    6, "Uninitiated",
    7, "Unreported",
    101, "Warning",
    102, "FIP",
    103, "Unimplemented",
    104, "Unapproved",
);

#----------------------------------------------------------------------------
# Prints time and date in a human-readable format
sub format_time {
    my $timestamp = shift;
    
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = 
        localtime ($timestamp);

    return sprintf ("%02d:%02d:%02d %02d-%02d-%04d\n", $hour, $min, $sec,
        $mday, ($mon + 1), ($year + 1900) );
}

sub html_entities {
    my $text = shift;
    
    $text =~ s/(^[\n\r\s]+|[\n\r\s]+$)//g;
    # Escape HTML entities
    $text =~ s/<br \/>/\n/gi;
    $text =~ s/&/&amp;/g;
    $text =~ s/</&lt;/g;
    $text =~ s/>/&gt;/g;
    $text =~ s/[\n\r]+/<br \/>\n/g;
    
    return $text;
}
#----------------------------------------------------------------------------

# Loads the problem database which is used for automatic identification
# of known test suite failures, warnings and false positives.
sub load_problem_database {
    my $file_name = shift;

    # Open the database file.
    open(FILE, $file_name) or fail("Could not open problem database $file_name");

    # Reset the line number counter.
    my $line_number = 0;

    # The problem record being read.
    my $problem;

    # Read the database line by line.
    my ($test_name, $test_point);
    while ($line = <FILE>) {

        # Increment the line number.
        $line_number++;

        # Skip empty lines and comment lines.
        next if ($line =~ /^\s*(\#.*)?$/);

        # The 'test' line starts a new test record.
        # It contains the name of the test and the test point number.
        if ($line =~ /^test\s+(.*)\s+(\d+)\s*$/) {

            # Store the test name and point.
            $test_name = $1;
            $test_point = $2;

            # Clear the old record and prepare a clean one.
            $problem = {
                test_name => $test_name,
                test_point => $test_point,
                markers => [],
                resolution => "",
                comment => "",
                quell => 0,
                lsb_report_id => "",
                lsb_resolution_id => "",
                expected_severity => ""
            };
        }

        # The 'expected_severity' line indicates the severity
        # expected to be reported by the test suite.
        elsif ($line =~ /^expected_severity\s*(.*)/) {
            my $expected_severity = $1;

            # Check if the expected severity is one of the predefined values.
            unless (($expected_severity eq "passed")
                || ($expected_severity eq "failed")
                || ($expected_severity eq "unresolved")
                || ($expected_severity eq "unused")
                || ($expected_severity eq "unsupported")
                || ($expected_severity eq "untested")
                || ($expected_severity eq "uninitiated")
                || ($expected_severity eq "unreported")
                || ($expected_severity eq "warning")
                || ($expected_severity eq "fip")
                || ($expected_severity eq "unimplemented")
                || ($expected_severity eq "unapproved")
            ) {
                fail("problem database: $test_name $test_point: " .
                    "unrecognized expected_severity '$1'");
            }

            $problem->{'expected_severity'} = $expected_severity;

        }

        # The 'marker' line contains a marker which can be found
        # to identify the particular problem from the messages.
        # There may be (and typically are) more markers, each on its own
        # 'marker' line.
        elsif ($line =~ /^marker\s*(.*)/) {
            push @{$problem->{'markers'}}, $1;
        }

        # The 'resolution' line contains the resolution of the problem.
        elsif ($line =~ /^resolution\s*(.*)/) {
            $problem->{'resolution'} = $1;
        }

        # If the 'quell' line is present, the problem is considered harmless
        # regardless of severity. This is used for already solved bugs
        # and false positives.
        elsif ($line =~ /^quell\s*(.*)/) {
            $problem->{'quell'} = 1;
        }

        # The 'comment' line holds a comment for human reader
        # which is copied to the problem report. There can be more
        # comment lines.
        elsif ($line =~ /^comment\s(.*)/) {
            $problem->{'comment'} .= "$1\n";
        }

        # The 'lsb_report_id' contains the LSB problem report ID.
        elsif ($line =~ /^lsb_report_id\s(\S*)/) {
            $problem->{'lsb_report_id'} = $1;
        }

        # The 'lsb_resolution_id' contains the LSB resolution ID.
        elsif ($line =~ /^lsb_resolution_id\s(\S*)/) {
            $problem->{'lsb_resolution_id'} = $1;
        }

        # The 'end' line finishes one problem entry.
        elsif ($line =~ /^end/) {

            # Check if this problem report has all required fields.
            if ($problem->{'expected_severity'} eq "") {

                # Expected severity must be provided.
                fail("problem database: $test_name $test_point: " .
                    "no expected_severity field");
            }

            # The problem record database is indexed by the test name
            # and point number, joined together to a key.
            # If there are multiple records for the same test point, they are
            # distinguished by an additional sequential number.
            my $record_key = $test_name . " " . $test_point;

            # If there are already some existing records for this test point,
            # find the next unused sequential number.
            my $sequential_number = 0;
            while (my $existing_record = $problem_database{$record_key}) {
                $sequential_number++;
                $record_key = $test_name . " " . $test_point . "($sequential_number)";
            }

            # Add the problem record to the database.
            $problem_database{$record_key} = $problem;
        }

        # Anything else is a syntax error.
        else {
            fail("problem database: syntax error at line $line_number");
        }
    }
    close FILE;
}

#----------------------------------------------------------------------------

# Finds a problem record in the database.
#
# Arguments:
# $1 - test name
# $2 - test point
# $3 - problem severity
# $4 - test messages
sub find_problem_record
{
    my ($test_name, $test_point, $severity, $test_messages) = @_;

    # Some automatically generated problem reports have no test name or point;
    # these are not meant to be searched in the database but directly reported
    # to the user.
    unless ($test_name) {
        # Do not bother searching, return nothing immediately.
        return undef;
    }

    # The basic record key is the name of the test and the test point number.
    my $record_key = $test_name . " " . $test_point;
    
    # Try finding the problem record.
    my $sequential_number = 0;
    while (my $problem_record = $problem_database{$record_key}) {

        # Check if the severity matches.
        my $expected_severity = $problem_record->{'expected_severity'};
        if (lc($severity) eq lc($expected_severity)) {

            # Look which distinguishing markers the problem should have.
            my $markers_ref = $problem_record->{'markers'};

            # If the problem record contains no markers, it matches only
            # if there were no messages.
            unless (@$markers_ref) {
                unless ($test_messages) {
                    return $problem_record;
                }
            }

            # Check if the markers are found in the test messages.
            my $markers_matched = 0;
            foreach (@$markers_ref) {
                if (index($test_messages, $_) >= 0) {
                    $markers_matched++;
                }
            }

            # All the markes must match.
            if ($markers_matched == @$markers_ref) {
                # Yup, this is the problem report we looked for.
                return $problem_record;
            }
        }

        # This report does not match, but there may be another one
        # for the same test point - increment the sequential number
        # and retry.
        $sequential_number++;

        # Add the sequential number to the record key.
        $record_key = $test_name . " " . $test_point . "($sequential_number)";
    }

    # Nothing was found.
    return undef;
}

#----------------------------------------------------------------------------

# Processes the result of a test point.
#
# Arguments:
# $1 - test number (0 if the tests are not numbered)
# $2 - test name
# $3 - test point (0 if the tests are not divided into test points)
# $4 - severity
# $5 - test messages
sub process_test_point_result {
    my ($test_number, $test_point, $test_name, $severity, $test_messages) = @_;

    # Sanity check: Test name must be valid.
    unless ($test_name) {
        debug_inform ("Empty test name");
        return;
    }

    # Compare severity strings in lowercase.
    my $lowercase_severity = lc($severity);

    # Increment the total test count.
    $stats_tests_total{$journal_title}++;

    # Passed and unused tests are not printed, only counted.
    if ($lowercase_severity eq "passed") {
        $stats_tests_passed{$journal_title}++;
        return;
    }
    elsif (($lowercase_severity eq "unused") || ($lowercase_severity eq "untested") ||
        ($lowercase_severity eq "unsupported") || ($lowercase_severity eq "unimplemented")) {
        $stats_tests_skipped{$journal_title}++;
        return;
    }

    my $quelled = 0;
    my $resolution = "";
    my $comment = "";
    my $lsb_report_id = "";
    my $lsb_resolution_id = "";

    # Compare the problem description with the database.
    my $problem_record = find_problem_record($test_name, $test_point, $severity, $test_messages);
    if ($problem_record) {

        # Load the data from the problem record.
        $quelled = $problem_record->{'quell'};
        $resolution = $problem_record->{'resolution'};
        $comment = $problem_record->{'comment'};
        $lsb_report_id = $problem_record->{'lsb_report_id'};
        $lsb_resolution_id = $problem_record->{'lsb_resolution_id'};

        # Add the "(quelled)" keyword to the severity if the problem is quelled.
        # This means it can be safely ignored.
        if ($quelled) {

            # Increment the quelled tests count statistic.
            $stats_tests_quelled{$journal_title}++;

            if ( $lsb_report_id ) {
                $severity .= " (waived)";
            } else {
                $severity .= " (quelled)";
            }
        }
    }
    
    # Print the problem to the table.
    print_problem_table_line(
        $test_number,
        $test_point,
        $test_name,
        $severity,
        $resolution,
        $comment,
        $lsb_report_id,
        $lsb_resolution_id,
        $test_messages
    );
}
#----------------------------------------------------------------------------

# Prints a header of the problem table. It is called automatically
# from report_problem() when the first problem is reported.
# (If there are no problems during the whole test run, the table
# is not printed at all.)
sub print_problem_table_header() {

    report("<h4>Problem Summary</h4>\n\n");
    report("<p>Click on lines in the table to see the details " .
        "about each problem.</p>\n\n");
    report("<table width=\"100%\">\n");
    report("<thead>\n");
    report("<tr style=\"font-weight:bold; background-color:$gray;\">\n");
    report("<td>Test Name</td><td>Severity</td><td>Resolution</td>\n");
    report("</tr>\n");
    report("</thead>\n");
    report("<tbody>\n");
}

#----------------------------------------------------------------------------

# Prints a line of the problem table.
#
# Arguments:
#
# $1 - test number (0 if the tests are not numbered, or if this problem line
#      is not connected to any particular test).
# $2 - test point number (0 if there is no test point).
# $3 - test name (empty if this problem line is not connected to any
#      particular test).
# $4 - severity.
# $5 - resolution (empty if unknown).
# $6 - comment
# $7 - report ID in the LSB database (empty if not yet reported).
# $8 - resolution ID in the LSB database (empty if not yet resolved).
# $9 - messages printed by the testing program.
sub print_problem_table_line
{
    my $test_number = shift;
    my $test_point = shift;
    my $test_name = shift;
    my $severity = shift;
    my $resolution = shift;
    my $comment = shift;
    my $lsb_report_id = shift;
    my $lsb_resolution_id = shift;
    my $test_messages = shift;

    # Compare the severity string case-insensitively (we are only humans).
    my $lowercase_severity = lc($severity);

    # Choose the color for the specified message severity.
    my $color;
    if ($lowercase_severity =~ /\(quelled|waived\)/) {
        # Anything what is quelled is harmless, so use neutral gray
        # for all quelled problems.
        $color = "gray";
    }
    elsif (($lowercase_severity eq "failed") || ($lowercase_severity eq "failure") ||
        ($lowercase_severity eq "unresolved") || ($lowercase_severity eq "unreported") ||
        ($lowercase_severity eq "uninitiated") || ($lowercase_severity eq "test suite error")
    ) {
        # Use fluorescent red for various failures.
        $color = "red";

        # Increment the number of failures reported.
        $stats_failures_reported{$journal_title}++;

        # Even a single failure in a single test is enough for flipping
        # the exit code from 0 to 1.
        $exit_code = 1 << 8;
    }
    elsif ($lowercase_severity eq "warning") {
        # Use fluorescent yellow for warnings.
        $color = "yellow";

        # Increment the number of warnings reported.
        $stats_warnings_reported{$journal_title}++;
    }
    elsif ($lowercase_severity eq "fip") {
        # Use fluorescent green for FIPs.
        $color = "green";
        
        $severity .= " [<a href='tests_help.pl#FIP' title='Help'>?</a>]";
        
        # Increment the number of FIPs reported.
        $stats_fips_reported{$journal_title}++;
    }
    elsif ($lowercase_severity eq "notice") {
        # Use neutral gray for harmless messages.
        $color = "gray";

        # Increment the number of notices reported.
        $stats_notices_reported{$journal_title}++;
    }
    else {
        debug_inform ("Unrecognized severity code '$severity'");
    }

    # If the problem table header was not yet printed, print it now.
    unless ($problem_table_printed) {
        print_problem_table_header();
        $problem_table_printed = 1;
    }

    # Give the message box a unique ID.
    my $message_box_id = "msgbox" . $msgbox_id; $msgbox_id++;

    # Start the line (with attached script that opens the detail block
    # on clicking).
    report("<tr class=\"ptl_$color\" title='Click here!' onclick=\"flip('$message_box_id');\">");

    # Print the test name and test point (if available).
    report("<td>");
    if ($test_name) {
        report("$test_name ");
    }
    else {
        report("(none) ");
    }
    if ($test_point > 0) {
        report("<span style=\"color:$dark_blue;\">$test_point</span>\n");
    }
    report("</td>\n");

    # Print severity and resolution.
    report("<td>$severity</td>\n");
    report("<td>$resolution</td>\n");

    # Finish the line.
    report("</tr>\n");

    # Print the detail block, which is initially hidden and unrolls when
    # the table row is clicked (via the javascript above).
    report("<tr id=\"$message_box_id\" style=\"display:none;\">\n");
    report("<td colspan=\"3\">");
    report("<div style='padding-left:32px;'>");
    report("Messages from the test\n");
    report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
    if ($test_messages) {
        report("<p style='font-family:monospace'>".$test_messages."</p>");
    } else {
        report("(none)");
    }
    report("\n</div>\n");
    if ($comment) {
        report("Comment:\n");
        report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
        report("$comment\n</div>\n");
    }
    if ($lsb_report_id or $lsb_resolution_id) {
        report("LSB Report Status:\n");
        report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
        if ($lsb_report_id) {
            report("Reported to the LSB Committee as ");
            report("<a href=\"https://www.linux-foundation.org/cert/pr.php?view=$lsb_report_id\">");
            report("PR $lsb_report_id</a>.<br />\n");
        }
        if ($lsb_resolution_id) {
            report("Resolved (resolution ID is ");
            report("<span style=\"color:$dark_blue;\">$lsb_resolution_id</span>).<br />\n");
        }
        report("</div>");
    }
    report("</div>");
    if ( $color eq "red" ) {
         my $j_name = $journal_name;
         $j_name =~ s/^.*\/([^\/]+)/$1/;
         my $esc_journal_title = $journal_title;
         $esc_journal_title =~ s/ /%20/g;
         my $esc_test_case = $test_name.' '.$test_point;
         $esc_test_case =~ s/ /%20/g;
         report("<div align=\"right\">"
            ."<input type=\"button\" value=\"Report bug in test\""
            ." onclick=\"javascript:bug_report('$final_results_dir/$results_subdir/lsb-autotest-$results_subdir.tgz', '$esc_journal_title', '$esc_test_case');\""
            ." title=\"Press this button if you beleive the problem is in the test, not in your system.\" /></div>"
            );
    }
    report("</td>");
    report("</tr>\n");
}

#----------------------------------------------------------------------------

# Prints the footer of the problem table. Does nothing if the problem table
# was not printed at all.
sub finish_problem_table() {

    # Complain if we have not seen any tests at all; this may mean the test suite
    # terminated somehow at the very start.
    if ($stats_tests_total{$journal_title} == 0) {
        print_problem_table_line(
            0,
            0,
            "(none)",
            "Failure",
            "No tests found in the journal",
            "",
            "",
            $line
        );
    }

    # Close the table or report that no problems were detected
    if ($problem_table_printed) {
        report("</tbody>\n");
        report("</table>\n");
    }
    else {
        report("<h4>Problem Summary</h4>\n");
        report("<p><font color='#006600'><b>No problems were detected.</b></font></p>\n");
    }
}

#----------------------------------------------------------------------------

# Starts a two-column table.
sub start_two_column_table {
    report("<table>\n");
    report("<tbody>\n");
}

#----------------------------------------------------------------------------

# Prints a single line of a two-column table.
#
# Arguments:
# $1 - The first column.
# $2 - The second column.
# $3 (optional) - The foreground color of the second column.
# $4 (optional) - The background color of the second column.
sub two_column_table_line {
    my $first_column = shift;
    my $second_column = shift;
    my $foreground_color = shift;
    my $background_color = shift;
    
    # Build the style specifier for the second column.
    my $second_column_style = "padding: 2px; margin: 2px;";
    
    # If the second column contains a number, print it right-aligned.
#    if ($second_column =~ /^[0-9\.]*$/) {
        $second_column_style .= "text-align: right; ";
#    }
    
    # Apply the foreground and background colors if specified.
    if ($foreground_color) {
        $second_column_style .= "color: $foreground_color; ";
    }
    if ($background_color) {
        $second_column_style .= "background-color: $background_color; ";
    }
    else {
        $second_column_style .= "background-color: $neutral_gray; ";
    }
    
    # Print the line with appropriate decorations.
    return "<tr style=\"padding: 4px; margin: 4px;\">".
    "<td style=\"background-color:$gray;\">$first_column</td>\n".
    "<td style=\"$second_column_style\">$second_column</td>\n".
    "</tr>\n";
}
#----------------------------------------------------------------------------

# Prints a two-column table line to the test report
sub print_two_column_table_line {
    report(two_column_table_line(@_));
}
#----------------------------------------------------------------------------

# Finishes a two-column table.
sub finish_two_column_table {
    report("</tbody>\n");
    report("</table>\n");
}

#----------------------------------------------------------------------------

sub print_statistics_line {
    my ($title, $value, $warn_if_nonzero, $warning_color) = @_;

    # Use neutral gray background by default.
    my $background_color = $neutral_gray;

    # If a warning is requested, change the color
    # when the value is nonzero.
    if ($warn_if_nonzero && ($value > 0) ) {
        print_two_column_table_line($title, $value, $dark_blue,
            $warning_color);
    }
    else {
        print_two_column_table_line($title, $value, $dark_blue);
    }
}


#----------------------------------------------------------------------------

# Prints the statistics.
sub print_statistics() {

    report("<h4>Statistics</h4>\n");

    # Calculate the time interval between the tests start and stop.
    my $test_duration = "unknown";
    if ( $test_object->{FINISH_TIME} && $test_object->{START_TIME} )
    {
        $test_duration = $test_object->{FINISH_TIME} - $test_object->{START_TIME};
        my $test_duration_sec = $test_duration % 60;
        $test_duration = int($test_duration / 60);
        my $test_duration_min = $test_duration % 60;
        my $test_duration_hour = int($test_duration / 60);
        
        if($test_duration_hour > 0) {
            $test_duration = "${test_duration_hour}h ${test_duration_min}m ${test_duration_sec}s";
        }
        else {
            $test_duration = "${test_duration_min}m:${test_duration_sec}s";
        }
    }
    
    # Print the statistics table.
    my $background_color;
    start_two_column_table();
    print_statistics_line("Test time", $test_duration, 0);
    print_statistics_line("Total tests", $stats_tests_total{$journal_title}, 0);
    if ($stats_tests_expected{$journal_title} > 0) {
        print_statistics_line("Total tests expected", $stats_tests_expected{$journal_title}, 
            $stats_tests_total{$journal_title}<$stats_tests_expected{$journal_title}?1:0, $fluorescent_yellow );
    }
    print_statistics_line("Passed tests", $stats_tests_passed{$journal_title}, 0);
    if ($stats_tests_skipped{$journal_title} > 0) {
        print_statistics_line("Tests not relevant/not applicable", $stats_tests_skipped{$journal_title}, 0);
    }
    if ($stats_tests_skips_unexpected{$journal_title} > 0) {
        print_statistics_line("Test skips unexpected", $stats_tests_skips_unexpected{$journal_title}, 1, $fluorescent_yellow);
    }
    if ($stats_tests_quelled{$journal_title} > 0) {
        print_statistics_line("Quelled problems", $stats_tests_quelled{$journal_title}, 0);
    }
    if ($stats_fips_reported{$journal_title} > 0) {
        print_statistics_line("FIPs", $stats_fips_reported{$journal_title}, 1, $fluorescent_green);
    }
    if ($stats_warnings_reported{$journal_title} > 0) {
        print_statistics_line("Warnings", $stats_warnings_reported{$journal_title}, 1, $fluorescent_yellow);
    }
    if ($stats_failures_reported{$journal_title} > 0) {
        print_statistics_line("Failures", $stats_failures_reported{$journal_title}, 1, $fluorescent_red);
    }
    
    # Give the final verdict.
    if ($stats_failures_reported{$journal_title} > 0) {
        print_two_column_table_line("Verdict", "Failure");
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        print_two_column_table_line("Verdict", "Warnings");
    }
    else {
        print_two_column_table_line("Verdict", "Success");
    }

    # Finish the table.
    finish_two_column_table();

    # Sanity check: If everything looks fine, check if the number of tests
    # matches the sum of skipped, passed, quelled and FIPed tests. This helps
    # preventing false negatives (which are most dangerous).
    if ($stats_failures_reported{$journal_title} == 0) {
        if ($stats_tests_skipped{$journal_title} + $stats_tests_passed{$journal_title} + $stats_fips_reported{$journal_title}
            + $stats_tests_quelled{$journal_title} != $stats_tests_total{$journal_title}) {

            # Oops.
            debug_inform ("No failures reported but some tests are missing");
        }
    }
}

#----------------------------------------------------------------------------

sub print_test_summary_line
{
    my ($title, $anchor) = @_;

    my $title_comment = sprintf ("%-25s", "<!-- $title -->");

    # Add the result into the single-line summary.
    $summary_line->print ("    $title_comment <td>");
    if ($stats_failures_reported{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"failure_icon.png\" alt=\"F\">");
        $summary_line->print ("</a>");
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"warning_icon.png\" alt=\"W\">");
        $summary_line->print ("</a>");
    }
    elsif ($stats_tests_total{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"ok_icon.png\" alt=\"OK\">");
        $summary_line->print ("</a>");
    }
    else {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"failure_icon.png\" alt=\"Un\"></a>");
    }
    $summary_line->print ("</td>\n");
}

#----------------------------------------------------------------------------

sub open_journal
{
    my $title = shift;
    my $anchor = shift;
    $journal_name = shift;
    $test_object = shift;

    # Store the journal title and anchor.
    $journal_title = $title;
    if ( $journal_title =~ /^([^\\\/]+)[\\\/]([^\\\/]+)[\\\/]([^\\\/]+)$/ ) {
        $journal_title = $3;
        $journal_group{$journal_title}{TOP_GROUP} = $1;
        $journal_group{$journal_title}{SUB_GROUP} = $2;
    } else {
        $journal_group{$journal_title}{TOP_GROUP} = "Unsorted";
        $journal_group{$journal_title}{SUB_GROUP} = "";
    }
    push @journal_titles, $journal_title;
    $journal_anchors{$journal_title} = $anchor;
    $test_versions{$journal_title} = $run_version{$test_object->name};
    
    # If journal isn't closed then close it.
    if ( $journal_file ) { close_journal (); }

    $problem_table_printed = 0;

    $stats_failures_reported{$journal_title} = 0;
    $stats_warnings_reported{$journal_title} = 0;
    $stats_fips_reported{$journal_title} = 0;
    $stats_notices_reported{$journal_title} = 0;
    $stats_tests_total{$journal_title} = 0;
    $stats_tests_passed{$journal_title} = 0;
    $stats_tests_skipped{$journal_title} = 0;
    $stats_tests_quelled{$journal_title} = 0;
    $stats_tests_expected{$journal_title} = 0;
    $stats_tests_skips_unexpected{$journal_title} = 0;
    
    debug_inform("Reading journal: $journal_name");
    # Open the journal file.
    $journal_file = new IO::File;
    unless (open ($journal_file, $journal_name)) {
        
        # The journal file could not be open; mention this in the log
        # (so that we know that something was missing) and return.
        
        $tests_cancelled++;
        
        report( "<p><font color='#A00000'><b>No result.</b></font>"
            ." $journal_title failed to run, interrupted"
            ." or crashed during the execution."
        );
        print_test_summary_line ($journal_title, $anchor);
        debug_inform( "Failed to read journal: $journal_name\."
            ."\n$journal_title failed to run, interrupted or crashed"
            ." during the execution.\n($journal_name was not present)."
        );
        $journal_file = undef;
    } else {
        if ($test_object->{DESCRIPTION}) {
            report("<p>".$test_object->{DESCRIPTION}."</p>");
        }
        report("<p><a href='get.pl$final_results_dir/$results_subdir/test_journals/"
                .extract_filename($journal_name)."'>Open the journal file</a></p>" );
    }
}

#----------------------------------------------------------------------------

sub close_journal
{
    unless ($journal_file) { return; }
    
    # Finish the problem table.
    finish_problem_table();

    # Print the statistics.
    print_statistics();

    # Add the result into the single-line summary.
    print_test_summary_line ( $journal_title, $journal_anchors{$journal_title} );

    # Close the journal file.
    $journal_file->close ();

    # Reset state variables.
    $journal_file = undef;
    $journal_title = "";
}

#----------------------------------------------------------------------------

# The main TET journal parser function which loads the TET journal
# and produces a HTML report. Uses stdin if the input file name is "-".
#
# Parameters:
#    $1 - the title to be printed before the report (typically the name
#         of the test the journal belongs to).
#    $2 - a HTML anchor to produce (useful for referencing to the results
#         of the test in a larger document).
#    $3 - the file name of the journal.
sub parse_tet_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }
    unless ($test_object) { return; }
    
    # This flag is set to nonzero if we are inside a test point.
    my $inside_test = 0;
    my $inside_test_point = 0;

    # Configuration variables (from lines with code 30)
    my %config_vars = ();

    # Read the journal line by line.
    my ($test_name, $test_number, $test_point, $test_messages);
    my $test_points_in_this_case;
    my $line_number = 0;
    while ($line = <$journal_file>) {
        $line_number++;
        # Each journal line consists of three parts, separated by
        # a vertical bar:
        #
        # 1. A numeric line code, determining the meaning of the line.
        # 2. Additional data, format depends on line code.
        # 3. Additional text (message for the user, or some comment).
        unless ($line =~ /^(\d*)\|(.*)\|(.*)$/) {

            # Lines that do not conform to this format are reported
            # to the user and skipped.
            print_problem_table_line(
                0,
                0,
                "(none)",
                "Warning",
                "Unrecognized entry in test journal",
                "$journal_name($line_number)",
                "",
                "",
                $line
            );
            next;
        }
        my $line_code = $1;
        my $line_data = $2;
        my $line_text = $3;

        # Determine the type of the line and process it appropriately.
        if ($line_code == 0) {
            # Code 0 -- Test Case Controller Start.
            # The line structure is:
            # 0|version time date|who
            unless ($line_data =~ /^(\S*)\s*(\S*)\s(\S*)$/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            # This line should be one of the first;
            # if it comes in the middle of normal results, it is strange.
            if ($problem_table_printed) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Test Case Start entry appears in the middle of the journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    "$line_data\n$line_text"
                );
            }
            else {
                my $test_engine_version = $1;
                my $test_start_time = $2;
                my $test_start_date = $3;
                my $test_command_line = $line_text;

                # Print the test engine version, time and date to the output.
                report("<h4>Test suite information</h4>");
                start_two_column_table();
                print_two_column_table_line("Test Engine Version", $test_engine_version);
                print_two_column_table_line("Test Started At", "$test_start_time $test_start_date");
                print_two_column_table_line("Test Command Line", "$test_command_line");
                finish_two_column_table();
            }
        }
        elsif ($line_code == 5) {

            # Code 5 -- Local System Information.
            # The line structure is:
            # 5|sysname nodename release #date machine|text
            unless ($line =~ /^5\|(\S*)\s*(\S*)\s*(\S*)\s*#(.*)\s(\S+)\|/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            # The system information should be one of the first items;
            # if it comes in the middle of normal results, it is strange.
            if ($problem_table_printed) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "System info entry appears in the middle of the journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    "$line_data\n$line_text"
                );
            }
            else {
                my $system_name = $1;
                my $node_name = $2;
                my $release = $3;
                my $date = $4;
                my $machine = $5;

                # Print the local system data.
                report("<h4>Tested system information</h4>\n");
                start_two_column_table();
                print_two_column_table_line("System Name", $system_name);
                print_two_column_table_line("Node Name", $node_name);
                print_two_column_table_line("Release", $release);
                print_two_column_table_line("Release Date", $date);
                print_two_column_table_line("Machine", $machine);
                finish_two_column_table();
            }
        }
        elsif ($line_code == 10) {

            # Code 10 -- Test Case Start.
            # The line structure is:
            # 10|activitynum testname hh:mm:ss|list of invocable components
            unless ($line =~ /^\d*\|(\d*)\s*(\S*)/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            if ($test_name) {
                # The previous test case was not properly finished;
                # issue a warning.
                print_problem_table_line(
                    $1,
                    0,
                    $test_name,
                    "Warning",
                    "Test case starts before the previous one ($test_name) has finished",
                    "$journal_name($line_number)",
                    "",
                    "",
                    ""
                );
            }

            # Store the test name and number.
            $test_number = $1;
            $test_name = $2;

            # Reset the test point number and counter.
            $test_point = 0;
            $test_points_in_this_case = 0;
        }
        elsif ($line_code == 15) {
            # Code 15 -- Test Case Manager Start.
            # Not interesting to us.
        }
        elsif ($line_code == 20) {
            # Code 20 -- Local/Remote System Configuration Start.
            # Not interesting to us.
        }
        elsif ($line_code == 30) {

            # Code 30 -- Configuration Variables.
            # The line structure is:
            # 30||VARIABLE=value
            if ($line =~ /^30\|\|([^=]+)=(.*)/) {
                my $varname = $1;
                my $value = $2;
                
                $config_vars{$varname} = $value;
            }
        }
        elsif ($line_code == 40) {
            # Code 40 -- Configuration End.
            
            # Print the local system data.
            report("<h4>Configuration information</h4>\n");
            start_two_column_table();
            for my $varname qw(VSX_NAME LSB_VERSION LSB_MODULES TET_VERSION XT_FONTPATH) {
                if ( exists $config_vars{$varname} ) {
                    print_two_column_table_line($varname, $config_vars{$varname});
                }
            }
            finish_two_column_table();
        }
        elsif ($line_code == 50) {
            # Code 50 -- Test Case Controller Message.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Test Suite Error",
                "Test case controller reports a problem",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 70) {
            # Code 70 -- Scenario Information.

            # LSB-runtime-test only:
            # for the .os sections need to double expected counts to allow
            # for the macro versions.
                if ( $line =~ /^70\|.*total tests in (ANSI|POSIX|LSB|PTHR).os ([0-9]+)/ ) {
                    $stats_tests_expected{$journal_title} += $2;
                    $stats_tests_expected{$journal_title} += $2;
                }
                elsif ( $line =~ /^70\|.*total tests in ([A-Za-z.0-9_-]+) ([0-9]+)/ ) {
                    $stats_tests_expected{$journal_title} += $2;
                }
                elsif ( $line =~ /total\s+tests\s+in/ ) {
                    complain ("$journal_name($line_number): A suspicious record in journal.")
                }
        }
        elsif ($line_code == 80) {
            # Code 80 -- Test Case End.
            unless ($test_name) {
                # Test case end found when no test case was started;
                # issue a warning.
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Test Case End entry found when no test case seems to be running",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $test_messages
                );
            }

            # Warn if the test case seems to be empty.
            if ($test_points_in_this_case == 0) {
                print_problem_table_line(
                    $test_number,
                    0,
                    $test_name,
                    "Warning",
                    "Test case contains no test points",
                    "",
                    "",
                    "",
                    ""
                );
            }

            # Clear the test name, number, and messages.
            $test_name = "";
            $test_number = 0;
            $test_messages = "";
            $test_points_in_this_case = 0;
        }
        elsif ($line_code == 90) {
            # Code 90 -- User Abort.
            print_problem_table_line(
                0,
                0,
                "(none)",
                "Test Suite Error",
                "Test suite execution interrupted by the user",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 100) {
            # Code 100 -- Captured Output.
            # Not interesting to us.
        }
        elsif ($line_code == 110) {
            # Code 110 -- Build Start.
            # Not interesting to us.
        }
        elsif ($line_code == 130) {
            # Code 130 -- Build End.
            # Not interesting to us.
        }
        elsif ($line_code == 200) {
            # Code 200 -- Test Purpose Start.
            # The structure of the line is as follows:
            # 200|activitynum pointnum hh:mm:ss|comment
            unless ($line =~ /^\d*\|(\d*)\s*(\d*)/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            # Check if the previous test point was properly finished.
            if ($test_point > 0) {
                if ($test_point == $2) {
                    # Some test suites have a strange habit of re-starting
                    # the same test point more than once.
                    print_problem_table_line(
                        $test_number,
                        $test_point,
                        $test_name,
                        "Notice",
                        "Test was restarted in the middle",
                        "",
                        "",
                        "",
                        ""
                    );
                }
                else {
                    # Notify about a disappeared test.
                    print_problem_table_line(
                        $test_number,
                        $2,
                        $test_name,
                        "Warning",
                        "Test point starts before the previous one ($test_point) has finished",
                        "",
                        "",
                        "",
                        ""
                    );
                }
            }

            # Clear the messages.
            $test_messages = "";

            # Store the test point number.
            $test_point = $2;

            # Increment the number of test points.
            $test_points_in_this_case++;
        }
        elsif ($line_code == 220) {
            # Code 220 -- Test Purpose Result.
            # The structure of the line is as follows:
            # 220|activitynum pointnum result time|comment
            unless ($line =~ /^\d*\|(\d*)\s*(\d*)\s*(\d*)/) {
                complain("Syntax error in journal");
                next;
            }

            # Print a warning if the test purpose seems to never been started.
            if ($test_point == 0) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Warning",
                    "Test result entry found but the test point had not started",
                    "",
                    "",
                    "",
                    $test_messages
                );
            }

            # Print a warning if the test purpose number is different from
            # what we expected.
            elsif ($test_point != $2) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Warning",
                    "Test purpose number at the end ($2) is different " .
                    "from the starting one ($test_point)",
                    "",
                    "",
                    "",
                    $test_messages
                );
            }

            # Get the result code.
            my $result_code = $3;

            # Find the meaning of the result code (severity).
            my $severity = $result_code_map{$result_code};
            unless ($severity) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Failure",
                    "Unrecognized test result code ($result_code)",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $test_messages
                );
            }
            else {
                # Process the result.
                process_test_point_result(
                    $test_number,
                    $test_point,
                    $test_name,
                    $severity,
                    $test_messages
                );
            }

            # Clear the test point number.
            $test_point = 0;
        }
        elsif ($line_code == 310) {
            # 310 -- Clean Start.
            # Not interesting to us.
        }
        elsif ($line_code == 320) {
            # 320 -- Clean End.
            # Not interesting to us.
        }
        elsif ($line_code == 400) {
            # 400 -- Invocable Component Start.
            # Not interesting to us.
        }
        elsif ($line_code == 410) {
            # 410 -- Invocable Component End.
            # Not interesting to us.
        }
        elsif ($line_code == 510) {
            # 510 -- Error reported by the underlying test suite.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Test Suite Error",
                "Error reported by the test suite",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 520) {
            # 520 -- Message from the underlying test suite.

            # Coalesce consequent rows of spaces in the message
            # to a single space. This improves matching with markers.
            my $message = $line_text;
            $message =~ s/\s+/ /g;

            # Add the message to the accumulated messages.
            $test_messages .= html_entities($message)."<br />";
        }
        elsif ($line_code == 600) {
            # 600 -- Parallel Start.
            # Not interesting to us.
        }
        elsif ($line_code == 620) {
            # 620 -- Parallel End.
            # Not interesting to us.
        }
        elsif ($line_code == 630) {

            # 630 -- Implied Sequential Start.
            # Not interesting to us.
        }
        elsif ($line_code == 640) {
            # 640 -- Implied Sequential End.
            # Not interesting to us.
        }
        elsif ($line_code == 700) {
            # 700 -- Repeat Start.
            # Not interesting to us.
        }
        elsif ($line_code == 720) {
            # 720 -- Repeat End.
            # Not interesting to us.
        }
        elsif ($line_code == 730) {
            # 730 -- Timed Loop Start.
            # Not interesting to us.
        }
        elsif ($line_code == 740) {
            # 740 -- Timed Loop End.
            # Not interesting to us.
        }
        elsif ($line_code == 750) {
            # 750 -- Random Start.
            # Not interesting to us.
        }
        elsif ($line_code == 760) {
            # 760 -- Random End.
            # Not interesting to us.
        }
        elsif ($line_code == 800) {
            # 800 -- Remote Start.
            # Not interesting to us.
        }
        elsif ($line_code == 820) {
            # 820 -- Remote End.
            # Not interesting to us.
        }
        elsif ($line_code == 830) {
            # 830 -- Distributed Start.
            # Not interesting to us.
        }
        elsif ($line_code == 840) {
            # 840 -- Distributed End.
            # Not interesting to us.
        }
        elsif ($line_code == 900) {
            # 900 -- Test Case Controller End.
            # Not interesting to us.
        }
        else {
            # Other lines are notified to the user and otherwise ignored.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Notice",
                "Unrecognized journal entry (code $line_code)",
                "$journal_name($line_number)",
                "",
                $line
            );
        }
    }

    # Finish the journal processing and close it.
    close_journal ();
}

#----------------------------------------------------------------------------

# The Tcl-style journal parser function which loads the journal
# and produces a HTML report. Uses stdin if the input file name is "-".
#
# Parameters:
#    $1 - the title to be printed before the report (typically the name
#         of the test the journal belongs to).
#    $2 - a HTML anchor to produce (useful for referencing to the results
#         of the test in a larger document).
#    $3 - the file name of the journal.
sub parse_tcl_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $test_number = 1;
    my ($test_name, $test_messages);
    while ($line = <$journal_file>) {

        # A test record message ends with a line with a rough format
        # '==== testname result'
        if ($line =~ /^=+\s*([a-zA-Z0-9_\-\.]+)\s*FAILED/) {

            unless ($test_name) {

                # Oops, we have missed something...
                print_problem_table_line(
                    $test_number,
                    0,
                    $1,
                    "Failure",
                    "End of test record found but no start was seen",
                    "",
                    "",
                    "",
                    $line
                );
            }

            # Process the result.
            process_test_point_result(
                $test_number,
                0,
                $test_name,
                "Failed",
                $test_messages
            );

            # Clear the previous test messages.
            $test_messages = "";

            # Maintain an artificial sequence of test numbers.
            $test_number++;

            # Finish the test record.
            $test_name = "";
        }

        # A failure report starts with a line with a rough format
        # '==== testname errormessage result'.
        elsif ($line =~ /^=+\s*([a-zA-Z0-9_\-\.]+)\s*(.+)\s*FAILED/) {

            # Load the test name.
            $test_name = $1;

            # Add the error message to the test messages.
            $test_messages .= html_entities($2)."<br />";
        }

        # A line that contains only the name of the test (which ends
        # with '.test', marks a passed test.
        elsif ($line =~ /^([a-zA-Z0-9_\-]+)\.test/) {
            process_test_point_result(
                $test_number,
                0,
                $1,
                "Passed",
                ""
            );

            # Maintain an artificial sequence of test numbers.
            $test_number++;
        }

        # Anything other?
        else {
            if ($test_name) {

                # We are inside a test point result record, so it will be
                # a test message.
                $test_messages .= html_entities($line)."<br />";
            }
            else {

                # Anything other is better ignored (there is too much noise
                # for considering unrecognized lines to be errors).
            }
        }
    }

    close_journal ();
}

#----------------------------------------------------------------------------

# The Python-style journal parser function which loads the journal
# and produces a HTML report. Uses stdin if the input file name is "-".
#
# Parameters:
#    $1 - the title to be printed before the report (typically the name
#         of the test the journal belongs to).
#    $2 - a HTML anchor to produce (useful for referencing to the results
#         of the test in a larger document).
#    $3 - the file name of the journal.
sub parse_python_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $inside_test_point = 0;
    my $test_number = 1;
    my ($test_name, $test_point, $test_messages);
    while ($line = <$journal_file>) {

        # Are we reading a test point record?
        if ($inside_test_point) {

            # Any line that resembles a start of another test record
            # stops this test record.
            if ($line =~ /^test_/) {
                $inside_test_point = 0;

                # Process this result.
                process_test_point_result(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Failure",
                    $test_messages
                );

                # Clear the previous test messages.
                $test_messages = "";
            }
            else {

                # Just append this line to the test record messages
                # and continue by the next line.
                $test_messages .= html_entities($line)."<br />";
                next;
            }
        }

        # Any other line is supposed to end the test point record.
        $inside_test_point = 0;

        # A line that contains only the name of the test (a single
        # identifier), marks a passed test.
        if ($line =~ /^([a-zA-Z0-9_]+)\s*$/) {
            process_test_point_result(
                $test_number,
                $test_point,
                $1,
                "Passed",
                ""
            );

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;
        }

        # A line in form 'testname skipped -- ...' marks a test
        # that was skipped, and the end of the line contains
        # the description why.
        elsif ($line =~ /^([a-zA-Z0-9_]+)\s*skipped\s*--\s*(.*)$/) {
            process_test_point_result(
                $test_number,
                $test_point,
                $1,
                "Unsupported",
                ""
            );

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;
        }

        # A line in form 'test testname failed -- ...' marks a failed test.
        elsif ($line =~ /^test\s*([a-zA-Z0-9_]+)\s*(failed|crashed)\s*--\s*(.*)$/) {

            # Get the test name.
            $test_name = $1;
            $test_point = 1;

            # We are reading the test failure report.
            $inside_test_point = 1;

            # Get the first line of the messages.
            $test_messages = html_entities($3);
        }
        
        elsif ($line =~ /^(\d+)\s+skips\s+unexpected/ ) {
            my $skips_unexpected = $1;
            $stats_tests_skips_unexpected{$journal_title} = $skips_unexpected;
        }
        
        # Other lines are silently ignored.
    }

    close_journal ();
}

#----------------------------------------------------------------------------

sub parse_simple_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $inside_test_point = 0;
    my $test_number = 1;
    my ($test_name, $test_point, $test_messages, $test_result);
    while ($line = <$journal_file>) {

        # Each line in form 'test name: result' should contain
        # the result of one test.
        if ($line =~ /^\s*([^:]*)\s*:\s*(\w*)/) {

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;

            # Load the name and result.
            $test_name = $1;
            $test_result = $2;

            # An optional detail message can follow.
            if ($line =~ /^\s*([^:]*)\s*:\s*(\w*)\s*:\s*(.*)$/) {
                $test_messages = $3;
            }

            # Check for failures.
            if ($test_result eq "failed") {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "failed",
                    $test_messages
                );
            }
            elsif ($test_result eq "OK") {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "passed",
                    $test_messages
                );
            }
            else {

                # Anything other is an error in the journal.
                print_problem_table_line(
                    $test_number,
                    0,
                    $test_name,
                    "failed",
                    "Syntax error in journal",
                    "",
                    "",
                    "",
                    $line
                );
            }
        }
        else {

            # Anything other is an error in the journal.
            print_problem_table_line(
                $test_number,
                0,
                $test_name,
                "failed",
                "Syntax error in journal",
                "",
                "",
                "",
                $line
            );
        }
    }

    close_journal ();
}
#----------------------------------------------------------------------------

# The Raptor-style journal parser function 
#
sub parse_raptor_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $test_number = 1;
    my ($test_name, $test_messages);
    while ($line = <$journal_file>) {
        
        chomp($line);
        
        if ($severity eq "FAILED") {
            if ($line !~ /^Checking\s/) {
                $test_messages .= html_entities($line)."<br />";
            } else {
                # End of previous ERROR message
                $test_messages = "Here is a diff between what expected and "
                    ."what obtained:\n".$test_messages;
                # Process the result.
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "FAILED",
                    $test_messages
                )
            }
        }
        
        # The journal starts with line
        # Testing legal rdf/xml
        if ($line =~ /^Testing /) {
            next;
        }

        # A test case looks like
        # Checking $test_name (ok|FAILED)
        if ($line =~ /^Checking\s+([^\s]+)\s+(ok|FAILED)$/)
        {
            $test_name = $1;
            my $result = $2;
            
            $test_messages = "";
            if ( $result eq "ok" ) {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "Passed",
                    ""
                );
                $severity = "";
            }
            elsif ( $result eq "FAILED" ) {
                $severity = "FAILED";
            }
            
            $test_number++;
        }
    }

    close_journal ();
}
#----------------------------------------------------------------------------


# Writes the information about the system and autotest to the output
# directory.
sub write_test_info
{
    # Write the host name, machine type, platform and OS version, start time.

    unless(-d "$results_dir/host_info") {
        cmd("mkdir -p $results_dir/host_info");
    }

    write_string_as_file ( "$results_dir/host_info/host_machine",
                           "$host_machine\n" );

    write_string_as_file ( "$results_dir/host_info/host_name", 
                           "$host_name\n" );

    write_string_as_file ( "$results_dir/host_info/host_platform",
                           "$host_platform\n" );

    write_string_as_file ( "$results_dir/host_info/host_os",
                           "$host_os\n" );

    write_string_as_file ( "$results_dir/host_info/host_kernel",
                           "$host_kernel\n" );

    write_string_as_file ( "$results_dir/host_info/etc_issue",
                           "$host_etc_issue\n" );

    # Write the versions of the installed lsb packages.
    append_string_to_file ( "$results_dir/host_info/package_versions", "LSB:\n"
      ."lsb\t".package_info("lsb")."\n"
      ."lsb-dtk-manager\t".package_info("lsb-dtk-manager")."\n"
    );

    unless(-d "$results_dir/autotest_info") {
        cmd("mkdir -p $results_dir/autotest_info");
    }

    write_string_as_file ( "$results_dir/autotest_info/start_time",
                           format_time($start_time) );

    write_string_as_file( "$results_dir/autotest_info/finish_time",
                          format_time($finish_time)  );

    unless(-d "$results_dir/test_logs") {
        cmd("mkdir -p $results_dir/test_logs");
    }
    
    # Copy the console log to test_logs
    cmd("cp -f /tmp/tests_output.8888 $results_dir/test_logs");
}
#----------------------------------------------------------------------------

sub report_index_line {
    my $journal_title = shift;
    my $display_journal_title = shift;
    
    # Cut the left spaces to bring it out of the hyperlink
    my $leftspace = "";
    if ($display_journal_title =~ /^(\s+)(.*)/ ) {
        $leftspace = $1;
        $display_journal_title = $2;
    }
  
    my $text = "";
    # Row starts
    $text .= "<tr style='background-color:".$neutral_gray."'>";
    
    # The first column: the hyperlink to the anchor in the report
    $text .= "<td>";
    $text .= "&nbsp;&nbsp;&nbsp;&nbsp;".$leftspace
            ."<b><a href='#".$journal_anchors{$journal_title}."'"
            ." title='Click here to proceed to details'>"
            .$display_journal_title."</a></b>";
    $text .= "</td>\n";
    
    # The second column: version of the testsuite
    $text .= "<td>";
    if ($journal_title !~ /::/ ) {
        $text .= "&nbsp; <font size='-1' color='#555555'>v "
            .$test_versions{$journal_title}."</font>";
    }
    $text .= "</td>\n";
    
    # The 3rd column: status
    my $bg_color = $neutral_gray;
    my $status = "";
    if ($stats_failures_reported{$journal_title} > 0) {
        $bg_color = $fluorescent_red;
        $status = "Failures: ".$stats_failures_reported{$journal_title};
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        $bg_color = $fluorescent_yellow;
        $status = "Warnings: ".$stats_warnings_reported{$journal_title};
    }
    elsif ($stats_tests_passed{$journal_title} == 0) {
        $bg_color = $fluorescent_red;
        $status = "<span title=\"'No result' means the tests failed"
        ." to run, interrupted or crashed during the execution.\">No Result"
        ." [<a href='tests_help.pl#No_result' title='Help'>?</a>]</span>";
    }
    elsif ($stats_tests_expected{$journal_title} > 0 
          && $stats_tests_expected{$journal_title} > $stats_tests_total{$journal_title} ) 
    {
        $bg_color = $fluorescent_yellow;
        $status = "Warning";
    }
    else {
        $bg_color = $fluorescent_green;
        $status = "Success";
    }
    # still the 3rd column: status
    $text .= "<td style=\"background-color:$bg_color\">$status";
    $text .= "</td>\n";
   
    # The 4th column: number of passed tests
    $text .= "<td>";
    $text .= "<span style='color:black'>Passed: ".
            $stats_tests_passed{$journal_title};
    $text .= "</span>";
    $text .= "</td>\n</tr>\n";
    return $text;
}

sub compile_report {
    # Print Title
    $report->print ("<h1>LSB automatic test results for $host_os on $host_platform</h1>\n");
    $report->print ("The complete console output log is available <a href=\"get.pl$final_results_dir/$results_subdir/test_logs/tests_output.8888\">here</a>.<br/><br/>"
    );
    $report->print ("<h3>Test Info</h3>\n");
    
    # Print general information
    $report->print (
    "<table>\n"
    .two_column_table_line("Test Started At", format_time($start_time))
    .two_column_table_line("Test Finished At", format_time($finish_time))
    .two_column_table_line("Host Name", $host_name)
    .two_column_table_line("Host OS", $host_os)
    .two_column_table_line("Host Machine", $host_machine)
    .two_column_table_line("Host Platform", $host_platform)
    .two_column_table_line("Host Kernel", $host_kernel)
#    .two_column_table_line("/etc/issue", "<pre>$host_etc_issue</pre>")
    ."</table>\n"
    );
    
    # Build index
    my %groups = ();
    foreach my $journal_title( @journal_titles ) {
        my $top_group = $journal_group{$journal_title}{TOP_GROUP};
        my $sub_group = $journal_group{$journal_title}{SUB_GROUP};
        if ( exists $groups{$top_group}{$sub_group} ) {
            push @{$groups{$top_group}{$sub_group}}, $journal_title;
        } else {
            my @ar = ($journal_title);
            $groups{$top_group}{$sub_group} = \@ar;
        }
    }
    
    my @journal_titles_sorted = ();
    my %subclasses = ();
    
    # Print the index
    $report->print("<h3>Tests executed</h3>");
    $report->print("<table border='0' cellpadding='2' cellspacing='1'>");
    
    foreach my $top_group ("Certification Tests", "Unofficial Tests", "Unsorted") {
        if ( exists $groups{$top_group} ) {
            $report->print( "<tr><td colspan='4' style='background-color:#b0b0ff'>"
                           ."&nbsp;<b>$top_group</b></td></tr>");
            foreach my $sub_group ("Command Tests", "Static Interface Tests", ""
                , "Runtime Interface Tests", "Automated Application Battery Tests"
            ) {
                if ( exists $groups{$top_group}{$sub_group}) {
                    my @sub_elems = @{$groups{$top_group}{$sub_group}};
                    $report->print( 
                         "<tr><td colspan='4' style='background-color:#d0d0f0'>"
                        ."&nbsp;&nbsp;&nbsp;<b>$sub_group</b></td></tr>" );
                    foreach my $journal_title (@sub_elems) {
                        my $display_journal_title = $journal_title;
                        if ( $journal_title =~ /(.*)::(.*)/ ) {
                            my $subclass = $1;
                            $display_journal_title = "&nbsp;&nbsp;&nbsp;&nbsp;".$2;
                            if ( ! exists $subclasses{$subclass} ) {
                                $report->print( "<tr style='background-color:#dbdbe8'><td>"
                                                ."&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<font color='#224466'><b>$subclass Tests:</b></font></td>"
                                                ."<td>&nbsp; <font size='-1' color='#555555'>v "
                                                .$test_versions{$journal_title}."</font></td>"
                                                ."<td colspan='2'></td></tr>"
                                               );
                                $subclasses{$subclass} = 1;
                            }
                        }
                        $report->print( report_index_line($journal_title, $display_journal_title) );
                        push @journal_titles_sorted, $journal_title;
                    }
                }
            }
        }
    }
    $report->print("</table>");
    
   
    # Test journals
    foreach $journal_title( @journal_titles_sorted ) {
        my $test_report = $reports{$journal_title};
        # Print the title.
        $report->print("<hr />\n");
        $report->print("<h2><a name=\"".$journal_anchors{$journal_title}
                ."\"></a><font color=\"#336699\">".$journal_title."</font> &nbsp; <font size='-1' color='#888888'>v ".$test_versions{$journal_title}."</font></h2>\n");
        
        $report->print ( $test_report );
    }
    
    # Print the host info to the one-line summary.
    $summary_line->print ("<table>\n <tr>\n");
    $summary_line->print ("    <!-- OS version -->       <td class=\"title\">$host_os</td>\n");
    $summary_line->print ("    <!-- Architecture -->     <td class=\"title\">$host_machine</td>\n");
    $summary_line->print (" </tr>\n</table>\n");
}

#----------------------------------------------------------------------------

sub sum {
    my $sum = 0;
    foreach (@_) { $sum += $_; }
    return $sum;
}

sub generate_INFO_file {
    my $INFO_file_name = shift;
    
    my $text = ""
        .$tests_total." Tests Total\n"
        .$tests_passed." Tests Passed\n"
        .$failures_reported." Failures\n"
        .$tests_quelled." Tests Quelled\n"
        .$tests_skipped." Tests Skipped\n"
        ;
    
    if ( $tests_cancelled > 0 ) {
        $text .= "There are cancelled tests\n";
    }
    
    $text .= "Verdict: $overall_verdict\n";
    
    $text .= "\n".$test_comments;
    
    write_string_as_file ($INFO_file_name, $text);
}
#----------------------------------------------------------------------------

sub pack_and_send_output
{
    #Go to the output directory
    chdir($results_dir);
    
    # Write INFO file
    unless ($omit_INFO ) {
        my $INFO_file_name = $output_file_name;
        $INFO_file_name =~ s/[^\/]*$/INFO/;
        generate_INFO_file ( $INFO_file_name );
    }
    
    # Create an output archive, and remove the temporary output directory.
    if ( $pack_files || $send_email) {
        inform ("Creating the output archive");
        my $files = capture("ls");
        cmd ("tar czf $output_file_name *");
        if ($remove_after_packing) {
            $files =~ s/[\n\r]+/ /g;
            cmd("rm -rf $files");
        }
    }
    
    if ( $send_email ) {
        # Send the results by e-mail if requested.
        debug_inform("Sending an e-mail");
        
        # Use sendmail utility
        &sendmail ( { 
            TO      => $tester_email,
            SUBJECT => "LSB autotest results for $host_name",
            TEXT    => "Please see the attached archive for the test results.",
            ATTACHMENT => "$output_file_name",
                    } );
            
        if (!$pack_files) {
            cmd ("rm -f $output_file_name");
        }
    }
    
    if ( $pack_files && $remove_after_packing ) {
        debug_inform ( "The output archive is ready: $final_results_dir/$results_subdir/$output_file_name" );
    }
}

sub Init {
    # Load the problem database (it should be in the same directory
    # as the main program, so simply get the program name and replace
    # the file part by 'problem_db').
    
    my $database_file_name = "$script_directory/problem_db";
    load_problem_database($database_file_name);
    
    # Open the report files.
    $report = new IO::File;
    unless ($report->open ("> $results_dir/test_log.html")) {
        fail ("Could not open the report file for writing");
    }
    $report->print ("<html>\n");
    $report->print ("
<head>
<style type=\"text/css\">
.ptl_gray { background: #E6E6E6;}
.ptl_gray:hover { background: #D6D6F0;}
.ptl_green { background: #CDFFCD;}
.ptl_green:hover { background: #A0FFA0;}
.ptl_yellow { background: #FFFFCD;}
.ptl_yellow:hover { background: #FFFF90;}
.ptl_red { background: #FFCDCD;}
.ptl_red:hover { background: #FFA0A0;}
.ptl_gray,.ptl_green,.ptl_yellow,.ptl_red { cursor: pointer; }
</style>
</head>
    ");
    $report->print ("<body>\n");
    $report->print ("
<script language=\"javascript\" type=\"text/javascript\">
// <![CDATA[
function flip (message_box_id) {
    msgbox = document.getElementById(message_box_id);
    if (msgbox.style.display != 'none') {
        msgbox.style.display = 'none';
    }
    else {
        msgbox.style.display = '';
    }
}
function bug_report(journal_path, test_suite, test_case) {
    if (confirm('A new E-mail to lf_lsbcert\@lists.linux-foundation.org will now be created.\\nPlease, specify all the necessary details and attach the journal files:\\n\\n' +
        journal_path + '\\n\\nContinue?'))
    {
        document.location = 'mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=Problem%20report:%20' + test_suite + '%20/%20' + test_case;
    }
}
// ]]>
</script>
            ");
        
    # Open the one-line-summary file.
    $summary_line = new IO::File;
    unless ($summary_line->open ("> $results_dir/one_line_summary.html")) {
        fail ("Could not open the one-line summary file for writing"
             ." ($results_dir/one_line_summary.html)" );
    }
    $summary_line->print ("<html>\n<body>\n");
}

#----------------------------------------------------------------------------

sub Finish {

    # Close the report files.
    $report->print ("</body>\n</html>\n");
    $report->close ();
    
    $summary_line->print ("</body>\n</html>\n");
    $summary_line->close ();

    return $exit_code;
}

#----------------------------------------------------------------------------

sub generate_certification_report {
    # Open the report files.
    my $c_report = new IO::File;
    unless ($c_report->open ("> $results_dir/cert.html")) {
        inform ("Could not open the report file for writing");
        return;
    }
    
    $c_report->print ("<html>\n<body>\n");

    my $email_body = "Body= Name: $tester_name;\n Organization: $organization;\n OS: $host_os;\n Architecture: $host_platform;\n LSB Version: $lsb_version";
    $email_body =~ s/ /%20/g;
    $email_body =~ s/[\r\n]/%0A/g;

    if ( $overall_verdict eq "Success" ) {
        $c_report->print ("<center>\n<h1 style=\"width: 70%; background-color: #00c000;"
                ." color: white; font-size: 150%;\">Congratulations!<br />\n"
                ."You have passed the LSB&nbsp;$lsb_version certification tests!<br /><a href=\"tests_results.pl?details=$results_subdir\""
                ." style=\"color: yellow;\"><u>View detailed report</u></a></h1>\n</center>\n\n"
        );
    }
    else
    {
        $c_report->print ("<center>\n<h1 style=\"width: 70%; background-color: #e00000;"
            ." color: white; font-size: 150%;\">Sorry, the certification"
            ." tests have failed.<br />Please, see the"
            ." <a href=\"tests_results.pl?details=$results_subdir\""
            ." style=\"color: yellow;\"><u>detailed report</u></a>"
            ." for specific problems.</h1>\n</center>\n\n"
        );
        $c_report->print(<<DATA);
<h2>Analyze the Failing Tests</h2>
<p align="justify">
Please analyze the <a href="tests_results.pl?details=$results_subdir"><b>detailed report</b></a>
to understand the reasons of each fail and classify them into the following groups:</p>
<ol>
<li><b>Confirmed FAILs</b>, which are due to the real inconsistencies of your system with the LSB standard. To continue with the LSB certification, you have to fix your implementation to remedy such FAILs and then rerun the tests again.</li>
<li><b>False FAILs</b>, which you believe are due to the incorrect tests.
Please report such FAILs to the <a href="mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=False%20positive%20FAIL&amp;$email_body">lf_lsbcert\@lists.linux-foundation.org</a> list.
If the problem reported is confirmed you will be granted a <i>waiver</i> so the test failure will not affect your ability to certify.</li>
<li><b>Unknown FAILs</b>, which you have no idea about. Please ask help at the <a href="mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=Unknown%20FAIL&amp;$email_body">lf_lsbcert\@lists.linux-foundation.org</a> list.</li>
</ol>
Note: only test results with all the FAILs waived are eligible for certification.
<br />
<br />

DATA
    }

    $c_report->print(<<DATA);
<h2>Locating Test Journals</h2>
You can find the consolidated test journals of this test run at the following path (click to download to another location):
<a href="get.pl$final_results_dir/$results_subdir/$output_file_name">$final_results_dir/$results_subdir/$output_file_name</a>.
You need to attach them when applying for certification or when communicating with the support staff.
<br />
<br />

<h2>Perform Manual Application Battery Tests</h2>
If you are certifying for the full "Core &amp; C++ &amp; Desktop" profile you also need to perform the 
manual <a href="http://www.linux-foundation.org/docs/lsbbook/appbat.html" target="_blank">
Application Battery</a> tests in addition to the automated tests already done:

<ol>
<li>Go to the 
    <a href="http://www.linux-foundation.org/download/#app-battery" target="_blank">
    Application Battery Download</a> page.
</li>
<li>Download the packages corresponding to the LSB version chosen.</li>
<li>Go to the <a href="http://www.linux-foundation.org/appbat/cert.html" target="_blank">
    LSB Functional Tests</a> page and perform the following interactive
    tests for the LSB version chosen (click particular test name there
    for instructions on how to do each test):
    <ul>
      <li><a href="http://www.linux-foundation.org/appbat/fvt-$lsb_version.0/lsb-celestia_fvt.html" target="_blank">lsb-celestia</a></li>
DATA

if ($lsb_version eq "3.1") {
    $c_report->print (<<DATA);
      <li><a href="http://www.linux-foundation.org/appbat/fvt-$lsb_version.0/lsb-gaim_fvt.html" target="_blank">lsb-gaim</a></li>
      <li><a href="http://www.linux-foundation.org/appbat/fvt-$lsb_version.0/lsb-leafpad_fvt.html" target="_blank">lsb-leafpad</a></li>
      <li><a href="http://www.linux-foundation.org/appbat/fvt-$lsb_version.0/lsb-scribus_fvt.html" target="_blank">lsb-scribus</a></li>
DATA
}

    $c_report->print (<<DATA);
      <li><a href="http://www.linux-foundation.org/appbat/fvt-$lsb_version.0/lsb-xpaint_fvt.html" target="_blank">lsb-xpaint</a></li>
      <li><a href="http://www.linux-foundation.org/appbat/fvt-$lsb_version.0/lsb-xpdf_fvt.html" target="_blank">lsb-xpdf</a> (<i>in this test you need to go through the steps 1 to 6 only</i>)</li>
    </ul>
</li>
</ol>
<br />

<h2>Apply for Certification</h2>
<p align="justify">With all the tests performed successfully, you may apply for certification. You will need to send a mail to <a href="mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=Certification%20test%20results&amp;$email_body">lf_lsbcert\@lists.linux-foundation.org</a> with the automated tests journals attached (see <i>Locating Test Journals</i> above)
and comments about the results of manual Application Battery Tests.</p>
<p align="justify">Organizational details of the certification procedure are described at the <a href="http://www.linux-foundation.org/en/Certification" target="_blank">Certification Home Page</a>.</p>

<h2>Viewing this Page Again</h2>
<p align="justify">This page has been saved as a part of the test run results. You may view it at any time by clicking particular test result at the <i>Results</i> page.</p>

</body>
</html>
DATA

    # Close the report files.
    $c_report->close ();
}

#----------------------------------------------------------------------------

# Generates a HTML report and detects test failures.
# Returns the return code of the reporting utility, which is 0 if all tests
# have passed, 1 if some tests have failed.
sub generate_report
{
    inform ("Generating HTML report");
    
    #Go to the output directory
    chdir($results_dir);
    
    system("for name in `ls *.tar.gz`; do "
            ."tar -xzf \$name --overwrite; "
            ."done;"
           );
    cmd("rm -rf *.tar.gz");
    
    # Write the test information.
    write_test_info ();
    
    # Open a report file and print a header
    Init();
    
    # Parse each test journal
    foreach my $test_name (@tests_to_run) 
    {
        $tests{$test_name}->journal ();
    }
    
    # Sum up statistics
    $failures_reported = sum( values %stats_failures_reported );
    $tests_total = sum( values %stats_tests_total );
    $tests_passed = sum( values %stats_tests_passed );
    $tests_quelled = sum( values %stats_tests_quelled );
    $tests_skipped = sum( values %stats_tests_skipped );
    
    # Give the final verdict
    if ( ($failures_reported > 0) || ($tests_cancelled > 0) ) {
        $overall_verdict = "Failed";
    }
    else {
        $overall_verdict = "Success";
    }

    compile_report ();
    
    if ( $certification_mode ) {
        generate_certification_report ();
    }
    
    my $exit_code = Finish();
    if ($exit_code == 0) {
        # Report was generated, no failures were detected.
        return 0;
    }
    elsif (($exit_code >> 8) == 1) {
        # Report was generated, some test failures were detected.
        return 1;
    }
    else {
        fail ("Error generating report (exit code $exit_code)");
    }
}

1; # return value   
