#!/usr/bin/perl -w 
# LSB Test Execution Framework
# Main executable script
#
# Copyright (C) 2007 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF with some parts
# derived from lsb-autotest written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use Fcntl ':flock';

# Is debugging mode (for developering purpose)
our $debugging = 0;

# verbose level: -v[0-2]
#   0 - minimal. Only error messages are put into stderr; 
#   1 - default; 
#   2 - extended debug information.
our $verbose_lvl = 1;

# The directory where the script itself is placed.
our $script_directory;

# Options specified on the command line.
our $tester_email = "";          # --mail-to
our $send_email = 0;
our $output_file_name = "";      # --output
our $certification_mode = 0;
    
our $keep_packages = 1;          # --keep-packages
my $cleanup = 0;                 # --cleanup
our $download_packages = 0;      # --download
our $try_beta_packages = 0;      # --beta
my $check_only = 0;              # --check-only
my $not_run = 0;                 # --not-run
my $print_help = 0;              # --help
my $print_list = 0;              # --list
our $list_refresh = 0;           # --list-refresh
our $pack_files = 0;             # --pack
our $remove_after_packing = 0;   # --compress
our $omit_INFO = 0;              # --omit-INFO

our $short_scenarios = 0;        # --short
    $short_scenarios = 1 if ($debugging);

# Local directory where packages are located.
our $package_dir = "";           # -d, --package-dir

our $profile_core = 0;           # -T, -Tcore,c++
our $profile_core_cmdchk = 0;    # -T, -Tcore,c++
our $profile_core_libchk = 0;    # -T, -Tcore,c++
our $libchk_qt4 = 0;             # libchk_qt4
our $desktop_qt4 = 0;            # desktop_qt4
    
# Properties of the host machine.
our $host_machine = "";
our $host_platform = "";
our $host_name = "";
our $host_os = "";
our $host_kernel = "";
our $host_triplet = "";
our $host_etc_issue = "";

# Components release version
our $lsb_version = "";

# Array of tests to run; they will be run in the specified order. 
our @tests_to_run;

# Test -> download mode
our %test_mode;

# Test names -> test objects.
our %tests;

# How the wget tool should be called.
# -r, --recursive         specify recursive download
# -nd, --no-directories   don't create directories
# -np, --no-parents       don't ascend to the parent directory.
# -w, --wait= SECONDS     wait SECONDS between retrievals
# --retr-symlinks         when recursing, get linked-to files, (not dir)
our $wget = "wget --timestamping -r -nd -np -w 3 --retr-symlinks --timeout=120 ";

# The time when the test was started.
our $start_time;
our $finish_time;

# Information about launch.
our $tester_name = "";
our $organization = "";
our $test_comments = "";

# Directory where reports will be temporarily stored
our $results_dir = "/tmp/lsb-dist-testkit-temp";
our $results_subdir = "";

my $default_final_results_dir = "/var/opt/lsb/test/manager/results";
my $default_package_directory = "/var/opt/lsb/test/manager/lsb-packages";

# Directory where all results and reports are finally stored.
our $final_results_dir = "";


# The directory where autotest is placed
our $autotest_dir = "./../autotest";

# A command to run autotest. Will be generated dynamically.
our $autotest_cmd = "";
#----------------------------------------------------------------------------

# Ensures that the specified path is absolute, prepending the current
# working directory to relative paths.
sub make_path_absolute
{
    my $input_path = shift;

    # Determine the current working directory.
    my $working_directory = `pwd`;
    chomp ($working_directory);

    # If the path is absolute (starts with a slash), leave it as it is.
    my $output_path;
    if ($input_path =~ /^\//) {
        $output_path = $input_path;
    }

    # If the path is relative (does not start with a slash), make it
    # absolute by prepending the current working directory.
    else {
        $output_path .= $working_directory . "/" . $input_path;
    }
    
    # Remove trailing "/."
    $output_path =~ s/(\/\.?)+$//;
    
    return $output_path;
}
#----------------------------------------------------------------------------

# Determines the directory the script is executed from,
# and stores it to the global 'script_directory' variable.
sub find_script_directory
{
    $script_directory = $0;
    $script_directory =~ s/[^\/]*$//; # Basename
    $script_directory = make_path_absolute ($script_directory);
    $autotest_dir = "$script_directory/../autotest";
}
#----------------------------------------------------------------------------

sub BEGIN {
    $| = 1; # Immediate flush of all output

    # LSB 3.1 / 16.4. Path For System Administration Utilities
    # Certain utilities used for system administration (and other privileged
    # commands) may be stored in /sbin, /usr/sbin, and /usr/local/sbin. 
    # Applications requiring to use commands identified as system 
    # administration utilities should add these directories to their PATH.
    $ENV{PATH} .= ":/sbin:/usr/sbin:/usr/local/sbin";

    # Append the script directory to the @INC to attach our own modules
    find_script_directory();
    push @INC, $script_directory;
}

use Common;
use Journal;
use Packages;
#=============================================================================
# Subroutines of the main script
#=============================================================================

sub search_for_test_modules
{
    my $ref = shift; # relative path
    my $dir = $script_directory."/Tests".$ref; 
    
    my @found_tests;
    
    # Open directory
    opendir DIR, $dir or die "Can't open directory $dir: $!\n";
    
    # Read directory for files and dirs
    my @files = readdir(DIR);
    
    foreach my $file(@files)
    {
        next if $file =~ /^\./; # skip hidden files, . and ..
        
        if (-f "$dir/$file")   # -f checks it is a file
        {
            # Search only for *_test.pm files.
            next if( $file !~ /^(.*_test)\.pm$/ );
            
            push @found_tests, "$ref/$1";
        }
        elsif (-d "$dir/$file")  # -d checks it is a directory
        {
            # Do recursive search
            push @found_tests, &search_for_test_modules("$ref/$file");
        }
    }
    closedir DIR;
    return @found_tests;
}

#----------------------------------------------------------------------------

# Prints a short help text.
sub print_help
{
    my @test_names;
    
    foreach(values %tests) {
        push @test_names, $_->name;
    }
    
    my $text = <<DATA;
    Usage: lsb-autotest [OPTION]... TESTNAME...
    Runs one or more of LSB tests and produces a HTML report.

    WARNING! PLEASE DO NOT USE THIS SCRIPT ON ANY MACHINE THAT HOLDS
    IMPORTANT DATA, OR PROVIDES IMPORTANT SERVICES! The tests
    sometimes put a heavy stress on the tested computer,
    and in rare unfortunate cases they can crash the system or damage
    the stored data. They also are a security risk as they pass with
    the root priviledges.

    Tests available:
        @test_names

    'all' means all available tests.
    'appbat' means app-battery tests.
    'runtime-tests' means runtime tests.
    'libchk_qt4' and 'desktop_qt4' are libchk and desktop tests with Qt4 checks.
        
    Additional options:

    -B, --beta               Use beta packages if available.
    --cleanup                Just uninstall test packages and exit.
    -d, --package-dir        Directory for the test packages.
    -r, --results-dir        Directory for results.
    -D, --download           Automatically download needed test packages.
    -h, --help               Print this help and exit.
    -m, --mail-to            Send the results to the e-mail address given.
    -o, --output             Output file name ('.tar.gz' will be appended).
    -p, --platform           Assume specified host platform.
    -t, --triplet            Assume specified host triplet.
    -pm                      Point out the proper package manager (RPM|dpkg)
    -v[0-2]                  Verbose level:
                             0 - only errors output,
                             1 - default output,
                             2 - extra output (for debug purposes).
    --check-only             Do some checks without installing and running 
                             any tests.
    --not-run                Install tests but do not run them.
    -f, --profile            Take the options from the profile.
    -c, --compress           Compress the results (pack and remove old files)
    --pack                   Create package with the results.
    -oI, --omit-INFO         Do not create INFO file.
    --lsb=<LSB_VERSION>      LSB version to test against.
    --short                  Use truncated scenarios for some tests 
                             (for debug and demonstration purposes).
    --cert                   Get through the certification process.
    -T, -Tcore,c++           This option is used to run tests against 
                             "Core and C++" profile (i.e. all "Desktop" tests 
                             will be omitted). It is supported starting from 
                             LSB version 3.1 update 1 and has no effect for 
                             earlier versions. '-Tcore,c++,desktop' is default,
                             it gives the same result as no -T.

    Please report any bugs and improvements to fsgteam\@ispras.ru.
DATA
    
    print( $text );
}
#----------------------------------------------------------------------------

# Load the command line options.
sub read_command_line
{
    my $package_dir_expected = 0;
    my $forced_platform_expected = 0;
    my $forced_triplet_expected = 0;
    my $forced_package_manager_expected = 0;
    my $tester_email_expected = 0;
    my $output_file_expected = 0;
    my $profile_expected = 0;
    my $results_dir_expected = 0;
    my $profile = "";

    foreach my $arg (@ARGV) {
        # Handle arguments for options which have one.

        if ($package_dir_expected) {
            $package_dir = $arg;
            $package_dir_expected = 0;
        }
        elsif ($forced_platform_expected) {
            $host_platform = $arg;
            $forced_platform_expected = 0;
            inform ("Platform manually set to: $host_platform");
        }
        elsif ($forced_triplet_expected) {
            $host_triplet = $arg;
            $forced_triplet_expected = 0;
            inform ("Host triplet manually set to: $host_triplet");
        }
        elsif ($forced_package_manager_expected) {
            $host_package_manager = $arg;
            $forced_package_manager_expected = 0;
            inform ("Package manager manually set to: $host_package_manager");
        }
        elsif ($tester_email_expected) {
            $tester_email = $arg;
            $tester_email_expected = 0;
            $send_email = 1;
        }
        elsif ($output_file_expected) {
            $output_file_name = $arg;
            $output_file_expected = 0;
        }
        elsif ($results_dir_expected) {
            $final_results_dir = $arg;
            $results_dir_expected = 0;
        }
        elsif ($profile_expected) {
            $profile = $arg;
            $profile_expected = 0;
            
            read_profile ($profile);
        }
        

        # Anything other is considered to be a command line option.

        else {
            if (($arg eq "-B") || ($arg eq "--beta")) {
                $try_beta_packages = 1;
            }
            elsif ($arg eq "--cleanup") {
                $cleanup = 1;
            }
            elsif (($arg eq "-d") || ($arg eq "--package-dir")) {
                $package_dir_expected = 1;
            }
            elsif ($arg =~ "--package-dir=(.*)") {
                $package_dir = $1;
                # Remove trailing "/."
                $package_dir =~ s/(\/\.?)+$//;
            }
            elsif (($arg eq "-r") || ($arg eq "--results-dir")) {
                $results_dir_expected = 1;
            }
            elsif ($arg =~ "--results-dir=(.*)") {
                $final_results_dir = $1;
                # Remove trailing "/."
                $final_results_dir =~ s/(\/\.?)+$//;
            }
            elsif (($arg eq "-D") || ($arg eq "--download")) {
                $download_packages = 1;
            }
            elsif (($arg eq "-h") || ($arg eq "--help")) {
                $print_help = 1;
            }
#            elsif (($arg eq "-k") || ($arg eq "--keep-packages") ) {
#                $keep_packages = 1;
#            }
            elsif (($arg eq "-m") || ($arg eq "--mail-to") || ($arg eq "--mailto")) {
                $tester_email_expected = 1;
            }
            elsif (($arg =~ "--mail-to=(.*)") || ($arg =~ "--mailto=(.*)")) {
                $tester_email = $1;
                $send_email = 1;
            }
            elsif (($arg eq "-o") || ($arg eq "--output")) {
                $output_file_expected = 1;
            }
            elsif ($arg =~ "--output=(.*)") {
                $output_file_name = $1;
            }
            elsif (($arg eq "-p") || ($arg eq "--platform")) {
                $forced_platform_expected = 1;
            }
            elsif ($arg =~ "--platform=(.*)") {
                $host_platform = $1;
                inform ("Platform manually set to: $host_platform");
            }
            elsif (($arg eq "-t") || ($arg eq "--triplet")) {
                $forced_triplet_expected = 1;
            }
            elsif ($arg =~ "--triplet=(.*)") {
                $host_triplet = $1;
                inform ("Host triplet manually set to: $host_triplet");
            }
            elsif ($arg eq "--pm") {
                $forced_package_manager_expected = 1;
                inform ("Package manager manually set to: $host_package_manager");
            }
            elsif ($arg =~ "--pm=(.*)") {
                $host_package_manager = $1;
            }
            elsif ($arg =~ "-v([0-2])") {
                $verbose_lvl = $1;
            }
            elsif ($arg =~ "--check-only") {
                $check_only = 1;
            }
            elsif ($arg =~ "--not-run") {
                $not_run = 1;
            }
            elsif ($arg eq "--list") {
                $print_list = 1;
            }
            elsif ($arg eq "--list-refresh") {
                $print_list = 1;
                $list_refresh = 1;
                $download_packages = 1;
            }
            elsif ($arg eq "--list-refresh-only") {
                $list_refresh = 1;
                $download_packages = 1;
            }
            elsif (($arg eq "-f") || ($arg eq "--profile")) {
                $profile_expected = 1;
            }
            elsif ($arg =~ "--profile=(.*)") {
                $profile = $1;
                
                read_profile ($profile);
            }
            elsif ( ($arg eq "-c") || ($arg eq "--compress") ) {
                $pack_files = 1;
                $remove_after_packing = 1;
            }
            elsif ( $arg eq "--pack" ) {
                $pack_files = 1;
            }
            elsif (($arg eq "-oI") || ($arg eq "--omit-INFO")) {
                $omit_INFO = 1;
            }
            elsif ($arg =~ "--lsb=(.*)") {
                $lsb_version = $1;
            }
            elsif ($arg eq "--short") {
                $short_scenarios = 1;
            }
            elsif ($arg eq "--cert") {
                $certification_mode = 1;
            }
            elsif (($arg eq "-T") || ($arg eq "-Tcore,c++")) {
                $profile_core = 1;
                $profile_core_cmdchk = 1;
                $profile_core_libchk = 1;
            }
            elsif ($arg eq "-Tcore,c++,desktop") {
                # Do nothing, this is default.
            }
            elsif ($arg eq "libchk_qt4") {
                $libchk_qt4 = 1;
                push @tests_to_run, "libchk";
            }
            elsif ($arg eq "desktop_qt4") {
                $desktop_qt4 = 1;
                push @tests_to_run, "desktop";
            }
            
            # 'all' means running all available tests (some tests may be
            # available only in beta).
            elsif ($arg eq "all") {
                # Use all released tests.
                foreach my $test_obj (values %tests) {
                    if ( $test_obj->is_available() ) {
                        push @tests_to_run, $test_obj->name;
                    }
                }
            }

            # 'appbat' means running the app battery tests.
            elsif ($arg eq "appbat") {
                foreach ("apache", "expect", "groff", "lynx", "python", 
                         "raptor", "rsync", "tcl", "xpdf", "samba") {
                    if ( !$tests{$_} ) {
                        fail_nomail("Unknown test name: $_ (Try '--help').");
                    }
                    if ( $tests{$_}->is_available() ) {
                        push (@tests_to_run, $_);
                    }
                }
            }
            elsif ($arg eq "runtime-tests") {
                foreach ("cmdchk", "libchk", "libstdc++", "runtime" )
                {
                    if ( $tests{$_}->is_available() ) {
                        push (@tests_to_run, $_);
                    }
                }
            }
            elsif ($arg eq "libstdcpp") {
                push (@tests_to_run, "libstdc++");
            }
            # These parameters determine which tests to run.
            elsif ( $tests{lc($arg)} )  { push (@tests_to_run, "$arg"); }

            else {

                # Invalid argument.
                fail_nomail ("Unrecognized argument $arg (Try '--help').");
            }
        }
    }

    # Check for missing argument at the end.

    if (
        $package_dir_expected or
        $forced_platform_expected or
        $forced_triplet_expected or
        $tester_email_expected or
        $forced_package_manager_expected or
        $output_file_expected or
        $profile_expected or
        $results_dir_expected
    ) {
        fail_nomail ("Missing commandline argument (Try '--help').");
    }
    
    if ( $certification_mode ) {
        # Check LSB version to test against.
        if ( $lsb_version !~ /^3.[01]$/ ) {
            fail_nomail ("Certification allowed only for LSB versions 3.0 and 3.1.");
        }
        
        $short_scenarios = 0;
            $short_scenarios = 1 if ($debugging);
        $try_beta_packages = 0;
        $download_packages = 1;
        
        if ($profile_core) {
            if ($libchk_qt4 || $desktop_qt4) {
                $libchk_qt4 = 0;
                $desktop_qt4 = 0;
                inform( "Warning: qt4 tests will be omitted"
                    ." because -T option is specified." );
            }
        }
        
        if ( $lsb_version =~ /^3.1$/ ) {
            if ($profile_core) {
                @tests_to_run = qw(cmdchk libchk libstdc++ runtime         vsw4 
                    groff rsync lynx expect apache python             tcl samba)
            }else{
                @tests_to_run = qw(cmdchk libchk libstdc++ runtime desktop vsw4 
                    groff rsync lynx expect apache python raptor xpdf tcl samba)
            }
        }
        elsif ( $lsb_version =~ /^3.0$/ ) {
            @tests_to_run = qw(cmdchk libchk libstdc++ runtime vsw4 
                groff rsync lynx expect apache python xpdf tcl samba)
        }
        
        # Overwrite @tests_to_run with the only 'cmdchk' test for debugging.
        @tests_to_run = ("cmdchk") if ($debugging);
        
        %test_mode = ();
    }
}
#----------------------------------------------------------------------------

# Does initial environment checks before anything else is done.
# Terminates the script if some of the checks fails.
sub do_initial_checks
{
    # Complain if we are not run as root.
    if ($> != 0 && !($print_list || $list_refresh) ) {
        fail_nomail ("This test must be run as root.");
    }

    # Check params
    if ( !$host_platform || $host_platform =~ "unknown" ) {
        fail_nomail("Couldn't recognize the host platform."
            ."\nSupported platforms are ".(join ", ", @all_platforms)."."
            ."\nYou can identify your platform using the command 'uname -i'."
        );
    }
    
    if(!$host_package_manager || $host_package_manager eq "unknown") {
        fail_nomail(
            "Couldn't recognize package manager for this distro ($host_os)."
            ."\nSupported package managers are $supported_package_managers."
        );
    }

    # Check for LSB components
    #if ( $host_package_manager eq "RPM" ) {
    #    run_command("rpm -q --whatprovides lsb-core-$host_platform");
    #}
    
    # Check platform-specific loader
    if ($lsb_version =~ /^(\d)/ ) {
        my $lsb_verison_major = $1;
        my $loader_name = "";
        if ($host_platform eq "ia32") {
            $loader_name = "/lib/ld-lsb.so.$lsb_verison_major";
        }elsif($host_platform eq "ia64") {
            $loader_name = "/lib/ld-lsb-ia64.so.$lsb_verison_major";
        }elsif($host_platform eq "ppc32") {
            $loader_name = "/lib/ld-lsb-ppc32.so.$lsb_verison_major";
        }elsif($host_platform eq "ppc64") {
            $loader_name = "/lib64/ld-lsb-ppc64.so.$lsb_verison_major";
        }elsif($host_platform eq "s390") {
            $loader_name = "/lib/ld-lsb-s390.so.$lsb_verison_major";
        }elsif($host_platform eq "s390x") {
            $loader_name = "/lib64/ld-lsb-s390x.so.$lsb_verison_major";
        }elsif($host_platform eq "amd64") {
            $loader_name = "/lib64/ld-lsb-x86-64.so.$lsb_verison_major";
        }
        if( system("ls $loader_name >/dev/null 2>&1") != 0) {
            fail("Couldn't find LSB loader ($loader_name)"
                ." for this platform ($host_platform)." 
                );
        }
    }
    
    # Check for basic utilities we will need.
    check_for_tool ("lsb_release");    
    check_for_tool ("passwd");
    check_for_tool ("uname");
    
    # wget is an optional tool for downloading packages.
    if($download_packages) { 
        # Look if the shell knows this tool.
        unless( does_shell_know ("wget") ) {
            fail_nomail ("Couldn't find a required tool: 'wget'."
            ."\nPackages can't be downloaded from internet without this tool."
            );
        }
    }
    
    # RPM is required for any package manager
    check_for_tool ("rpm");
    if ( $host_package_manager eq "dpkg") {
        # Check for alien and dpkg
        check_for_tool("alien");
        check_for_tool("dpkg");
    }
    
    if ( $send_email ) {
        check_for_tool("/usr/sbin/sendmail");
    }
    
    # Tests custom checks.
    foreach my $test_name (@tests_to_run) 
    {
        my $test_obj = $tests{$test_name};
        debug_inform($test_obj->name." check...");
        $test_obj->check();
    }
}

#----------------------------------------------------------------------------

# This function creates objects, which work with test suites.
sub create_test_objects {
    
    my @test_files = &search_for_test_modules("");

    foreach my $test_name (@test_files) {
        $test_name = "\u$test_name";
        my $test_module_name = $test_name;
        $test_module_name =~ s/\//::/g;
        $test_module_name = "Tests::$test_module_name";
        $test_name = extract_filename($test_name);
        
        my $test_obj;
        
        # Link up the module and create object.
        eval "use $test_module_name; \$test_obj = $test_name->new();";
        if ($@) { # Any errors?
            complain($@);
            exit(2);
        }
        
        # Store created objects to a hash
        $tests{$test_obj->name} = $test_obj;
    }
}
#----------------------------------------------------------------------------

# Removes previously installed test packages.
sub purge_packages {

    # Collect all package names to a set, then purge it
    my %packages_to_purge;
    my @packages_to_purge;

    foreach my $test_name (@tests_to_run) 
    {
        my $test_obj = $tests{$test_name};
        foreach my $package_name ( split(" ",$test_obj->packages))
        { 
            if(!$packages_to_purge{$package_name})
            {
                $packages_to_purge{$package_name} = 1;
                push @packages_to_purge, $package_name;
            }
        
        }
        
        #$test_obj->purge();
    }
    
    # Deinstall in reverse order
    @packages_to_purge = reverse @packages_to_purge;
    
    my $num_to_uninstall;
    
    # If removing fails, e. g. due to dependencies, then retry
    # until number of packages stops to decrease.
    do {
        $num_to_uninstall = @packages_to_purge;
        
        for (my $i = 0; $i<$num_to_uninstall; $i++)
        {
            my $package_to_purge = shift @packages_to_purge;

            inform ("Purging package $package_to_purge...");
            
            # Uninstalls the specified package if it is installed.
            remove_package($package_to_purge);
            
            # Check if the package was really uninstalled
            if ( !package_info($package_to_purge) )
            {
                inform ("Ok.");
            }
            else
            {
                inform ("Fail.");
                
                # Retry to uninstall the packages, which haven't uninstalled properly
                push @packages_to_purge, $package_to_purge;
            }
        }
        # Retry to uninstall the packages, which haven't uninstalled properly
    } while (@packages_to_purge < $num_to_uninstall);
}
#----------------------------------------------------------------------------

# Reads information about the host machine.
sub read_host_info
{
    # Check for basic utilities we will need here.
    check_for_tool ("lsb_release");
    if ( capture ("lsb_release 2>&1") =~ /No LSB modules/ ) {
    	fail("No LSB modules are available.");
    }
    
    check_for_tool ("uname");
    
    # The machine type and architecture is discovered via uname.
    $host_machine = `uname -m`; chomp ($host_machine);
    $host_name = `uname -n`; chomp ($host_name);
    
    # Discover the host platform.
        
    # If uname utility supports -i options then use it
    if (system ("uname --help | grep -- ' -i\\W' >/dev/null") == 0)
    {
        $host_platform = `uname -i`; chomp ($host_platform);
        $host_platform = guess_platform($host_platform);
    }
        
    if ( !$host_platform || $host_platform =~ "unknown" )
    {
        # Try to find out the host platform using the host_machine.
        $host_platform = guess_platform($host_machine);
    }
        
    $host_kernel = `uname -r -v`; chomp ($host_kernel);

    # The OS version is determined via the lsb_release tool
    # (this should be the canonical, distro independent way).
    $host_os = `lsb_release -s -d`;
    $host_os =~ s/\"(.*)\"/$1/;    # remove quotes
    chomp($host_os);               # remove trailing whitespace
    
    if ( !$lsb_version ) {
        # Trying to detect supported LSB version
        my $lsb_release_v = `lsb_release -s -v`;
        if    ( $lsb_release_v =~ /4\.0/ ) { $lsb_version = "4.0"; }
        elsif ( $lsb_release_v =~ /3\.2/ ) { $lsb_version = "3.2"; }
        elsif ( $lsb_release_v =~ /3\.1/ ) { $lsb_version = "3.1"; }
        elsif ( $lsb_release_v =~ /3\.0/ ) { $lsb_version = "3.0"; }
        elsif ( $lsb_release_v =~ /2\.0/ ) { $lsb_version = "2.0"; }
        else { $lsb_version = "3.1"; }
    }
    # Guess the host description triplet.
    $host_triplet = "$host_machine\-gnu-linux";

    # Read the /etc/issue file, if present. Sometimes it is the only place
    # where an exact OS version can be found.
    $host_etc_issue = `cat /etc/issue`;

    $host_package_manager = guess_package_manager();
}
#----------------------------------------------------------------------------

# Tries to set the package source to the specified directory, provided
# that no package source was set before, and the directory exists.
sub try_package_directory
{
    my $directory_name = shift;

    # Do nothing if the package source is already set.
    if ($package_dir) {
        return;
    }

    # Try if the directory exists, if it is, use it.
    if (-d $directory_name) {
        $package_dir = $directory_name;
        inform ("Using package directory: $directory_name");
    }
}

#----------------------------------------------------------------------------

# Finds where the packages are (or should be) stored.
sub prepare_package_directory
{
    if ($package_dir) {

        # Ensure that the package dir path is absolute, so that
        # we do not depend on the working directory.
        $package_dir = make_path_absolute($package_dir);
    }
    else {

        # If no package source is set, look for some default locations.
        try_package_directory ($default_package_directory);
        
        # Not found?
        unless ($package_dir) {

            # Create directory for packages in a default place
            $package_dir = $default_package_directory;
            inform ("Using package directory: $package_dir");
            unless (-d $package_dir) { cmd ("mkdir -p \"$package_dir\""); }

            # Otherwise there is no way to get the packages, so stop.
            unless (-d $package_dir) {
                fail_nomail ("Can't create package directory $package_dir"
                   ." (Do you have enough rights?)."
                );
            }
        }
    }
}
#----------------------------------------------------------------------------

# Does all test-dependent cleanups
sub tests_cleanup 
{
    foreach my $test_name (@tests_to_run) 
    {
        if($tests{$test_name})
            { $tests{$test_name}->cleanup (); }
    }
}
#----------------------------------------------------------------------------

# Does all necessary cleanups at the end of the autotest run.
# This must be called at the program termination, otherwise we are leaving
# the system in unstable and insecure state.
sub cleanup
{
    debug_inform("-- Cleanup:");
    
    tests_cleanup();
}
#----------------------------------------------------------------------------

sub read_profile {
    my $profile = shift;
    
    open (FILE, $profile) or fail("Can't open profile $profile");
    
    while (my $line = <FILE> )
    {
        chomp($line);
        
        if( $line =~ /^\s*NAME:\s+(.*)$/ ) {
            $tester_name = $1;
        }
        elsif( $line =~ /^\s*ORGANIZATION:\s+(.*)$/ ) {
            $organization = $1;
        }
        elsif( $line =~ /^\s*EMAIL:\s+(.*)$/ ) {
            $tester_email = $1;
        }
        elsif( $line =~ /^\s*SEND_EMAIL:\s+([0-1])\s*$/ ) {
            $send_email = $1;
        }
        elsif( $line =~ /^\s*VERBOSE_LEVEL:\s+([0-2])\s*$/ ) {
            $verbose_lvl = $1;
        }
        elsif( $line =~ /^\s*KEEP_PACKAGES:\s+([0-1])\s*$/ ) {
            $keep_packages = $1;
        }
        elsif( $line =~ /^\s*ARCHITECTURE:\s+(.*)$/ ) {
            $host_platform = $1;
        }
        elsif( $line =~ /^\s*USE_INTERNET:\s+([0-1])\s*$/ ) {
            $download_packages = $1;
        }        
        elsif( $line =~ /^\s*COMMENTS:\s+(.*)$/ ) {
            $test_comments .= "$1\n";
        }
        elsif( $line =~ /^\s*SHORT_SCENARIOS:\s+(.*)$/ ) {
            $short_scenarios = $1;
        }
        elsif( $line =~ /^\s*LSB_VERSION:\s+(.*)$/ ) {
            $lsb_version = $1;
        }
        elsif( $line =~ /^\s*CERTIFY:\s+(.*)$/ ) {
            $certification_mode = $1;
        }
        elsif( $line =~ /^\s*PACKAGE_DIR:\s+(.*)$/ ) {
            $package_dir = $1;
        }
        elsif( $line =~ /^\s*RESULTS_DIR:\s+(.*)$/ ) {
            $final_results_dir = $1;
        }
        elsif( $line =~ /^\s*OUTPUT_FILE_NAME:\s+(.*)$/ ) {
            $output_file_name = $1;
        }
        elsif( $line =~ /^\s*PLATFORM:\s+(.*)$/ ) {
            $host_platform = $1;
        }
        elsif( $line =~ /^\s*PACKAGE_MANAGER:\s+(.*)$/ ) {
            $host_package_manager = $1;
        }
        elsif( $line =~ /^\s*LSB_PROFILE:\s+(.*)$/ ) {
            my $lsb_profile = $1;
            if ( $lsb_profile =~ /^core$/ ) {
                $profile_core = 1;
                $profile_core_cmdchk = 1;
                $profile_core_libchk = 1;
            }
        }
        elsif ( $line =~ /^\s*(\S*)\s+([pab])\s*$/ ) {
            my $test_name = $1;
            my $mode = $2;
            
            $test_name =~ s/libstdcpp/libstdc++/;
            
            push @tests_to_run, $test_name;
            
            $test_mode{$test_name} = $mode;
        }
        elsif ( $line =~ /^\s*cmdchk-opt\s+(\d+)\s*$/ ) {
            my $cmdchk_opt = $1;
            if ($cmdchk_opt == 2) {
                $profile_core_cmdchk = 1;
            } else {
                # Do nothing, default;
            }
        }
        elsif ( $line =~ /^\s*libchk-opt\s+(\d+)\s*$/ ) {
            my $libchk_opt = $1;
            if ($libchk_opt == 2) {
                $profile_core_libchk = 1;
            } 
            elsif ($libchk_opt == 3) {
                # Do nothing, default;
            }
            elsif ($libchk_opt == 4) {
                $libchk_qt4 = 1;
            }
        }
        elsif ( $line =~ /^\s*desktop-opt\s+(\d+)\s*$/ ) {
            my $desktop_opt = $1;
            if ( 4 == $desktop_opt ) {
                $desktop_qt4 = 1;
            }
        }
        else {
            #fail_nomail( "Syntax error in profile $profile;\nLine: $.: $line" );
        }
    }
    
    close (FILE);
}

#============================================================================
# The main script
#============================================================================

# Find our own directory. 
find_script_directory ();

# Create test objects according to @tests_to_run list.
create_test_objects ();

# Read information about the host machine.
read_host_info ();

# Read the command line.
read_command_line ();

# if '--help' switch is used, print help and exit.
if($print_help)
{
    &print_help();
    exit(0);
}

# Set Wget verbose level
if ($verbose_lvl == 0) 
    { $wget .= " -q"; }

# Setup Carp module.
if ($verbose_lvl >= 2)
    { $SIG{__DIE__} = sub { require Carp; Carp::confess(@_) } };

inform ("Script directory is $script_directory");

# Finds where the packages are (or should be) stored.
prepare_package_directory ();

if($print_list || $list_refresh) {
    @tests_to_run = keys %tests;
    &list_files ("b");
    if ($print_list) {
        &print_files_list ();
    }
    exit(0);
}

inform ("Cmd line: @ARGV");
inform ("Host name: $host_name");
inform ("Host machine: $host_machine");
inform ("Host platform: $host_platform");
inform ("Host kernel: $host_kernel");
inform ("Host OS: $host_os");
inform ("Host triplet: $host_triplet");
inform ("Host package manager: $host_package_manager");
inform ("LSB version: $lsb_version");

# If the '--cleanup' switch is used, just purge all packages and temporary
# files, then exit.
if ($cleanup) {

    inform ("Purging test packages");
    purge_packages ();

    inform ("Purging temporary files");
    cmd ("rm -rf /tmp/lsb-autotest-*");

    inform ("Finished");
    exit (0);
}

# Look for available packages
inform("Looking for available packages");
inquire_for_files_list ();

if (@tests_to_run) {
    inform ("Tests to be run: @tests_to_run");
} else {
    fail_nomail ("No tests were chosen (use 'all' to execute all tests).");
}

# Do some initial environment checking.
do_initial_checks();

# Check for autotest
unless (   (-x "$autotest_dir/lsb-autotest") ) {
    fail("Can't find lsb-autotest in $autotest_dir");
}

# Create a clean directory for results
create_empty_directory($results_dir);
create_empty_directory("$results_dir/autotest_info");
create_empty_directory("$results_dir/host_info");
create_empty_directory("$results_dir/test_journals");
create_empty_directory("$results_dir/test_logs");

# Get the time when the tests were started.
$start_time = time ();

if ($check_only) { exit(0); }

# RUN

foreach my $test_name (@tests_to_run) 
{
    my $test_obj = $tests{$test_name};
    
    inform("Prepare $test_name test");
    my $result = prepare_test ($test_obj);
    
    if ($result != 0) {
        inform ("Error: Failed to prepare test $test_name."
                ." This test will be skipped."
        );
        next;
    }
    
    unless ( $not_run ) {
        # Remember start time
        $test_obj->{START_TIME} = time ();

        # Compile command line for the autotest.
        my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday)
            = localtime ( $test_obj->{START_TIME} );
        
        my $autotest_output_file_name = sprintf(
            "%s-%s-%s-%04d-%02d-%02d-%02dh-%02dm-%02ds",
            $host_platform, $host_name, $test_name, 
            $year + 1900, $mon + 1, $mday, $hour, $min, $sec
        );

        $autotest_cmd = 
            "$autotest_dir/lsb-autotest -v --trust-packages --keep-packages --beta --no-download".
            " -p \"$host_platform\" -o \"$results_dir/$autotest_output_file_name\""
            .( ($test_obj->name eq "cmdchk") && $profile_core_cmdchk?" -T":"")
            .( ($test_obj->name eq "libchk") && $profile_core_libchk?" -T":"")
            .( ($test_obj->name eq "libchk") && $libchk_qt4?" --qt4":"")
            .( ($test_obj->name eq "desktop") && $desktop_qt4?" --qt4":"")
            ;

        # Run the test
        inform("Run $test_name test");
        $result = $tests{$test_name}->run ();
        
        # Remember finish time
        $test_obj->{FINISH_TIME} = time ();

        debug_inform("$test_name test has returned code $result");
        if( $result == 111 ) {
            inform("Stop tests execution");
            last;
        }
    }
            
    unless ($keep_packages) {

        # Uninstall test packages.
        # This ensures that all config files will be fresh on the next
        # start and also that all packages really belong to the selected 
        # platform (on multi-platform machines).
        inform ("Purging old packages");
        purge_packages ();
    }
}

if ($not_run) { exit(0); }

# Get the time when all the tests were finished.
$finish_time = time();

# move results from temporary dir to 'results' dir.
unless( $final_results_dir ) {
    $final_results_dir = $default_final_results_dir;
}
unless( -d $final_results_dir) {
    cmd("mkdir -p '$final_results_dir'");
}
my($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = localtime ($start_time);
$results_subdir = sprintf( "%s-%s-%04d-%02d-%02d-%02dh-%02dm-%02ds",
  $host_platform, $host_name, $year + 1900, $mon + 1, $mday, $hour, $min, $sec );

debug_inform ("Result directory is $final_results_dir/$results_subdir");

# New reporter. Disabled yet.
#cmd ("if cd $results_dir; then $autotest_dir/lsb-autotest-reporter .;fi");
###

# If the output file is not specified, generate a name using
# the host machine name and the current time.
unless ($output_file_name) {

    # Get the starting time and break it up to components.
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = localtime ($start_time);

    # Generate the name for the output directory and file using
    # the host machine type, host name, and the test start time.
    $output_file_name = sprintf(
        "lsb-autotest-%s-%s-%04d-%02d-%02d-%02dh-%02dm-%02ds.tgz",
        $host_platform, $host_name,
        $year + 1900, $mon + 1, $mday, $hour, $min, $sec );
}
if ( $output_file_name !~ /\./ ) {
    $output_file_name .= ".tgz";
}    

# Generate a HTML report and detect test failures.
# Pack the output and optionally send it by e-mail.
my $test_failures_detected = generate_report ();

# Pack the output and optionally send it by e-mail.
pack_and_send_output();

debug_inform("Moving the result directory.");

run_command("rm -rf \"$final_results_dir/$results_subdir\"");
run_command("mv -f \"$results_dir\" \"$final_results_dir/$results_subdir\"");

inform("Results have been placed into $final_results_dir/$results_subdir");

# Make a record in the HISTORY file
open(FILE, ">> $final_results_dir/HISTORY");
flock(FILE, LOCK_EX);
print FILE "$results_subdir\n";
flock(FILE, LOCK_UN);
close(FILE);

inform ("Finished");

# Finish and always return 0
exit (0);

sub END
{
    my $errcode = $?;
    
    if ($error_msg) {
        my $tef_directory = $script_directory;
        $tef_directory =~ s/\/[^\/]*$//;
        my $message = 
            "#----------------------------------------------------------------"
            ."\n# There were some error messages during the execution."
            ."\n# You could read the file $tef_directory/README,"
            ."\n# 'Troubleshooting' section to get more information about them."
            ."\n# If you can't find a suitable solution,"
            ."\n# please report the bug to [fsgteam\@ispras.ru].";

        if($verbose_lvl > 0) {
            print( "$message\n" );
        }
            
    }
    
    # Remove all temporary changes
    unless ($print_list || $list_refresh) {
        cleanup();
    }
    
    $? = $errcode;
}
